"""Import functionality"""

import logging
import sys
from datetime import datetime
from typing import Iterable, List, Optional

import openpyxl
from openpyxl.workbook import Workbook
from openpyxl.worksheet.worksheet import Worksheet

from cg.exc import CgError
from cg.store import Store, models

from .models import ApplicationSchema, ApplicationVersionSchema, ParsedApplicationVersion

LOG = logging.getLogger(__name__)


def import_application_versions(
    store: Store, excel_path: str, sign: str, dry_run: bool, skip_missing: bool
):
    """
    Imports all application versions from the specified excel file
    Args:
        :param store:               Status database store
        :param excel_path:          Path to excel file with headers: 'App tag', 'Valid from',
                                    'Standard', 'Priority', 'Express', 'Research'
        :param sign:                Signature of user running the script
        :param dry_run:             Test run, no changes to the database
        :param skip_missing:        Continue despite missing applications
    """
    application_versions: Iterable[ParsedApplicationVersion] = parse_application_versions(
        excel_path=excel_path
    )

    for application_version in application_versions:
        application_obj: models.Application = store.application(application_version.app_tag)

        if not application_obj:
            LOG.error(
                "Failed to find application! Please manually add application: %s",
                application_version.app_tag,
            )

            if skip_missing:
                continue

            LOG.error("Rolling back transaction.")
            store.rollback()
            sys.exit()

        app_tag: str = application_obj.tag
        latest_version: models.ApplicationVersion = store.latest_version(
            application_version.app_tag
        )

        if latest_version and versions_are_same(
            version_obj=latest_version, application_version=application_version
        ):
            LOG.info("skipping redundant application version for app tag %s", app_tag)
            continue

        LOG.info("adding new application version to transaction for app tag %s", app_tag)
        new_version = add_application_version(
            application_obj=application_obj,
            latest_version=latest_version,
            version=application_version,
            sign=sign,
            store=store,
        )
        store.add(new_version)

    if not dry_run:
        LOG.info(
            "all application versions successfully added to transaction, committing " "transaction"
        )
        store.commit()
    else:
        LOG.error("Dry-run, rolling back transaction.")
        store.rollback()


def import_applications(
    store: Store, excel_path: str, sign: str, dry_run: bool, sheet_name: Optional[str] = None
):
    """
    Imports all applications from the specified excel file
    Args:
        :param sheet_name:          Optional sheet name
        :param store:               status database store
        :param excel_path:          Path to excel file with headers: 'App tag', 'Valid from',
                                    'Standard', 'Priority', 'Express', 'Research'
        :param sign:                Signature of user running the script
        :param dry_run:             Test run, no changes to the database
    """
    applications: Iterable[ApplicationSchema] = parse_applications(
        excel_path=excel_path, sheet_name=sheet_name
    )

    for application in applications:
        application_obj: models.Application = store.application(application.tag)
        if application_obj and applications_are_same(
            application_obj=application_obj, application=application
        ):
            LOG.info("skipping redundant application %s", application.tag)
            continue

        LOG.info("adding new application to transaction %s", application.tag)
        new_application = add_application_object(application=application, sign=sign, store=store)
        store.add(new_application)

    if not dry_run:
        LOG.info("all applications successfully added to transaction, committing transaction")
        store.commit()
    else:
        LOG.error("Dry-run, rolling back transaction.")
        store.rollback()


def prices_are_same(first_price: float, second_price: float) -> bool:
    """Checks if the given prices are to be considered equal"""

    if first_price == second_price:
        return True
    if first_price and second_price:
        return int(round(float(first_price))) == int(round(float(second_price)))

    return False


def versions_are_same(
    version_obj: models.ApplicationVersion, application_version: ApplicationVersionSchema
) -> bool:
    """Checks if the given versions are to be considered equal"""
    return (
        version_obj.application.tag == application_version.app_tag
        and version_obj.valid_from == application_version.valid_from
        and prices_are_same(version_obj.price_standard, application_version.standard)
        and prices_are_same(version_obj.price_priority, application_version.priority)
        and prices_are_same(version_obj.price_express, application_version.express)
        and prices_are_same(version_obj.price_research, application_version.research)
    )


def applications_are_same(
    application_obj: models.Application, application: ApplicationSchema
) -> bool:
    """Checks if the given applications are to be considered equal"""

    return application_obj.tag == application.tag


def add_application_version(
    application_obj: models.Application,
    latest_version: Optional[models.ApplicationVersion],
    version: ApplicationVersionSchema,
    sign: str,
    store: Store,
) -> models.ApplicationVersion:
    new_version = store.add_version(
        application=application_obj,
        version=latest_version.version + 1 if latest_version else 1,
        valid_from=version.valid_from,
        prices={
            "standard": version.standard,
            "priority": version.priority,
            "express": version.express,
            "research": version.research,
        },
        comment="Added by %s" % sign,
    )
    return new_version


def add_application_object(
    application: ApplicationSchema, sign: str, store: Store
) -> models.Application:
    """Adds an application from a raw application record"""
    new_application: models.Application = store.add_application(
        tag=application.tag,
        category=application.prep_category,
        description=application.description,
        is_accredited=application.is_accredited == 1.0,
        turnaround_time=application.turnaround_time,
        minimum_order=application.minimum_order,
        sequencing_depth=application.sequencing_depth,
        target_reads=application.target_reads,
        sample_amount=application.sample_amount,
        sample_volume=application.sample_volume,
        sample_concentration=application.sample_concentration,
        priority_processing=application.priority_processing == 1.0,
        details=application.details,
        limitations=application.limitations,
        percent_kth=application.percent_kth,
        percent_reads_guaranteed=application.percent_reads_guaranteed,
        comment=application.comment + " Added by %s" % sign,
        is_archived=application.is_archived == 1.0,
        is_external=application.is_external == 1.0,
        created_at=datetime.now(),
    )
    return new_application


def import_apptags(
    store: Store,
    excel_path: str,
    prep_category: str,
    signature: str,
    sheet_name: str,
    tag_column: int,
    tag_row: int,
    activate: bool,
    inactivate: bool,
):
    """Syncs all applications from the specified excel file"""

    orderform_application_tags = []

    for tag in get_cells_from_excel(
        excel_path=excel_path, sheet_name=sheet_name, tag_column=tag_column, tag_row=tag_row
    ):
        LOG.info("Found: %s in orderform", tag)
        orderform_application_tags.append(tag)

    if not orderform_application_tags:
        message = f"No applications found in column {tag_column} (zero-based), exiting"
        LOG.error(message)
        return

    for orderform_application_tag in orderform_application_tags:
        application_obj = store.application(tag=orderform_application_tag)

        if not application_obj:
            message = f"Application {orderform_application_tag} was not found"
            LOG.error(message)
            return

        if application_obj.prep_category != prep_category:
            message = (
                f"{orderform_application_tag} prep_category, expected: {prep_category} was:"
                f" {application_obj.prep_category}"
            )
            LOG.error(message)
            return

        if application_obj.is_archived:
            if activate:
                application_obj.comment = (
                    f"{application_obj.comment}\n{str(datetime.now())[:-10]}"
                    f"Application un-archived by {signature}"
                )
                application_obj.is_archived = False
                LOG.info("Un-archiving %s", application_obj)
            else:
                LOG.warning(
                    "%s is marked as archived but is used in the orderform, consider "
                    "activating it",
                    application_obj,
                )
        else:
            LOG.info("%s is already active, no need to activate it", application_obj)

    all_active_apps_for_category = store.applications(category=prep_category, archived=False)

    for active_application in all_active_apps_for_category:
        if active_application.tag in orderform_application_tags:
            LOG.info("%s was found in orderform tags, no need to archive it", active_application)
            continue
        if not inactivate:
            LOG.warning(
                "%s is marked as active but is not used in the orderform, consider archiving it",
                active_application,
            )
            continue
        active_application.is_archived = True
        active_application.comment = (
            f"{active_application.comment}"
            f"\n{str(datetime.now())[:-10]} "
            f"Application archived by {signature}"
        )
        LOG.info("Archiving %s", active_application)

    if not activate and not inactivate:
        LOG.info("no change mode requested, rolling back transaction")
        store.rollback()
    else:
        LOG.info("all applications successfully synced, committing transaction")
        store.commit()


def get_cells_from_excel(
    excel_path: str, sheet_name: str = "Drop down list", tag_column: int = 2, tag_row: int = 0
) -> Iterable[str]:
    workbook: Workbook = openpyxl.load_workbook(filename=excel_path, read_only=True, data_only=True)
    data_sheet: Worksheet = workbook[sheet_name]

    for row_idx, row in enumerate(data_sheet.rows):
        if row_idx < tag_row:
            continue
        value = row[tag_column].value
        if value is None:
            continue
        yield value


def parse_application_versions(
    excel_path: str, sheet_name: Optional[str] = None
) -> Iterable[ParsedApplicationVersion]:
    workbook: Workbook = openpyxl.load_workbook(filename=excel_path, read_only=True, data_only=True)
    data_sheet = workbook.worksheets[0]
    if sheet_name:
        data_sheet = workbook[sheet_name]

    application_info: dict
    for application_info in get_raw_dicts_from_xlsheet(data_sheet):
        yield ParsedApplicationVersion(**application_info)


def parse_applications(
    excel_path: str, sheet_name: Optional[str] = None
) -> Iterable[ApplicationSchema]:
    workbook: Workbook = openpyxl.load_workbook(filename=excel_path, read_only=True, data_only=True)
    data_sheet = workbook.worksheets[0]
    if sheet_name:
        data_sheet = workbook[sheet_name]

    application_info: dict
    for application_info in get_raw_dicts_from_xlsheet(data_sheet):
        yield ApplicationSchema(**application_info)


def get_raw_dicts_from_xlsheet(sheet: Worksheet) -> List[dict]:
    """Get the relevant rows from a sheet."""
    raw_data = []
    header_row = []
    first_row = True

    for row in sheet.rows:
        if row[0].value is None:
            break

        if first_row:
            header_row = [cell.value for cell in row if cell.value is not None]
            first_row = False
        else:
            values = [cell.value for cell in row]
            version_dict = dict(zip(header_row, values))
            raw_data.append(version_dict)

    return raw_data



"""Module that handles data and information regarding flow cells and flow cell directories. Used to
inspect and clean flow cell directories in demultiplexed runs."""
import logging
import re
import shutil
from pathlib import Path
from typing import List, Optional

from alchy import Query
from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.tb import TrailblazerAPI
from cg.constants import FlowCellStatus
from cg.constants.demultiplexing import DemultiplexingDirsAndFiles
from cg.constants.sequencing import Sequencers, sequencer_types
from cg.constants.symbols import ASTERISK
from cg.constants.housekeeper_tags import SequencingFileTag
from cg.store import Store

FLOW_CELL_IDENTIFIER_POSITION = 3
LOG = logging.getLogger(__name__)
SEQUENCER_IDENTIFIER_POSITION = 1


class DemultiplexedRunsFlowCell:
    """Class to check if a given flow cell in demultiplexed-runs is valid, or can be removed. A
    valid flow cell is named correctly, has the correct status ('ondisk') in Statusdb,
    and has FASTQ and/or SPRING files in Housekeeper.

    A flow cell that is correctly named has four segments divided by underscores:
    <run_date>_<sequencer>_<run_number>_<identifier>. The identifier consists of the position of
    the flow cell on the sequencer in the first position (either A or B), and the flow cell id in
    the rest of the identifier. Flow cell id is the naming convention used in Statusdb and other
    systems"""

    def __init__(
        self,
        flow_cell_path: Path,
        status_db: Store,
        housekeeper_api: HousekeeperAPI,
        trailblazer_api: Optional[TrailblazerAPI] = None,
        sample_sheets_dir: Optional[str] = None,
        fastq_files: Optional[Query] = None,
        spring_files: Optional[Query] = None,
    ):
        self.sample_sheets_dir: Path = Path(sample_sheets_dir) if sample_sheets_dir else None
        self.path: Path = flow_cell_path
        self.status_db: Store = status_db
        self.hk: HousekeeperAPI = housekeeper_api
        self.tb: TrailblazerAPI = trailblazer_api
        self.all_fastq_files: Optional[Query] = fastq_files
        self.all_spring_files: Optional[Query] = spring_files
        self.run_name: str = self.path.name
        self.split_name: List[str] = re.split("[_.]", self.run_name)
        self.identifier: str = self.split_name[FLOW_CELL_IDENTIFIER_POSITION]
        self.sequencer_serial_number: str = self.split_name[SEQUENCER_IDENTIFIER_POSITION]
        self.id: str = self.identifier[1:]
        self._hk_fastq_files: Optional[list] = None
        self._hk_spring_files: Optional[list] = None
        self._is_correctly_named: Optional[bool] = None
        self._exists_in_statusdb: Optional[bool] = None
        self._fastq_files_exist_in_housekeeper: Optional[bool] = None
        self._spring_files_exist_in_housekeeper: Optional[bool] = None
        self._files_exist_in_housekeeper: Optional[bool] = None
        self._files_exist_on_disk: Optional[bool] = None
        self._passed_check: Optional[bool] = None
        self._sequencer_type: Optional[str] = None
        self._is_demultiplexing_ongoing_or_started_and_not_completed: Optional[bool] = None

    @property
    def sequencer_type(self) -> str:
        """The type of sequencer a flow cell has been sequenced on."""
        if self._sequencer_type is None:
            self._sequencer_type: str = sequencer_types.get(
                self.sequencer_serial_number, Sequencers.OTHER
            )
        return self._sequencer_type

    @property
    def hk_fastq_files(self) -> list:
        """All FASTQ files in Housekeeper for a particular flow cell."""
        if self._hk_fastq_files is None:
            self._hk_fastq_files = [
                fastq_file for fastq_file in self.all_fastq_files if self.id in fastq_file.path
            ]
        return self._hk_fastq_files

    @property
    def hk_spring_files(self) -> list:
        """All spring files in Housekeeper for a particular flow cell"""
        if self._hk_spring_files is None:
            self._hk_spring_files = [
                spring_file for spring_file in self.all_spring_files if self.id in spring_file.path
            ]
        return self._hk_spring_files

    @property
    def is_correctly_named(self) -> bool:
        """Checks if the flow cell directory is correctly named: it should end with the flow cell
        id."""
        if self._is_correctly_named is None:
            self._is_correctly_named: bool = self.split_name[-1] == self.identifier
            if not self._is_correctly_named:
                LOG.warning(f"Flow cell not correctly named!: {self.path}")
        return self._is_correctly_named

    @property
    def exists_in_statusdb(self) -> bool:
        """Checks if flow cell exists in Statusdb."""
        if self._exists_in_statusdb is None:
            self._exists_in_statusdb: bool = self.status_db.get_flow_cell(self.id) is not None
            if not self._exists_in_statusdb:
                LOG.warning(f"Flow cell {self.id} does not exist in Statusdb!")
        return self._exists_in_statusdb

    @property
    def fastq_files_exist_in_housekeeper(self) -> bool:
        """Checks if the flow cell has any FASTQ files in Housekeeper."""
        if self._fastq_files_exist_in_housekeeper is None:
            self._fastq_files_exist_in_housekeeper: bool = len(self.hk_fastq_files) > 0
            if not self._fastq_files_exist_in_housekeeper:
                LOG.warning(f"Flow cell {self.id} does not have any FASTQ files in Housekeeper!")
        return self._fastq_files_exist_in_housekeeper

    @property
    def spring_files_exist_in_housekeeper(self) -> bool:
        """Checks if the flow cell has any SPRING files in Housekeeper."""
        if self._spring_files_exist_in_housekeeper is None:
            self._spring_files_exist_in_housekeeper: bool = len(self.hk_spring_files) > 0
            if not self._spring_files_exist_in_housekeeper:
                LOG.warning(f"Flow cell {self.id} does not have any SPRING files in Housekeeper!")
        return self._spring_files_exist_in_housekeeper

    @property
    def files_exist_in_housekeeper(self) -> bool:
        """Checks if the flow cell has any files, either FASTQ or SPRING, in Housekeeper."""
        if self._files_exist_in_housekeeper is None:
            self._files_exist_in_housekeeper: bool = (
                self.fastq_files_exist_in_housekeeper or self.spring_files_exist_in_housekeeper
            )
            if not self._files_exist_in_housekeeper:
                LOG.warning(
                    f"Flow cell {self.id} has neither FASTQ files nor SPRING files in Housekeeper!"
                )
        return self._files_exist_in_housekeeper

    @staticmethod
    def _check_files_existence(files: list) -> List[bool]:
        """Checks file existence and handles permission errors."""
        files_exist: List[bool] = []
        for file in files:
            try:
                files_exist.append(Path(file.path).exists())
            except PermissionError:
                LOG.warning(f"Can't check file {file}, no permission!")
                continue
        return files_exist

    @property
    def files_exist_on_disk(self) -> bool:
        """Checks if the SPRING or FASTQ files that are in Housekeeper are actually present on
        disk."""

        if self._files_exist_on_disk is None:
            if self.files_exist_in_housekeeper:
                if self.fastq_files_exist_in_housekeeper:
                    fastq_files_exist_on_disk: List[bool] = self._check_files_existence(
                        self.hk_fastq_files
                    )
                    self._files_exist_on_disk: bool = all(fastq_files_exist_on_disk)
                if self.spring_files_exist_in_housekeeper:
                    spring_files_exist_on_disk: List[bool] = self._check_files_existence(
                        self.hk_spring_files
                    )
                    self._files_exist_on_disk: bool = all(spring_files_exist_on_disk)
                if not self._files_exist_on_disk:
                    LOG.warning(f"Flow cell {self.id} has no fastq files or spring files on disk!")

            else:
                LOG.warning(
                    f"Flow cell {self.id} has no FASTQ files or SPRING files in Housekeeper, skipping "
                    "disk check"
                )
                self._files_exist_on_disk = False
        return self._files_exist_on_disk

    @property
    def is_demultiplexing_ongoing_or_started_and_not_completed(self) -> bool:
        """Checks demultiplexing status for a flow cell."""
        if self._is_demultiplexing_ongoing_or_started_and_not_completed is None:
            LOG.info(f"Checking demultiplexing status for flow cell {self.id}:")
            self._is_demultiplexing_ongoing_or_started_and_not_completed: bool = (
                self.tb.has_latest_analysis_started(case_id=self.id)
                or self.tb.is_latest_analysis_ongoing(case_id=self.id)
            ) and not self.tb.is_latest_analysis_completed(case_id=self.id)
            if self._is_demultiplexing_ongoing_or_started_and_not_completed:
                LOG.warning(f"Demultiplexing not fully completed for flow cell {self.id}!")
            else:
                LOG.info("Demultiplexing fully completed!")
        return self._is_demultiplexing_ongoing_or_started_and_not_completed

    @property
    def passed_check(self) -> bool:
        """Indicates if all checks have passed."""
        if self._passed_check is None:
            LOG.info(f"Checking {self.path}:")
            self._passed_check = all(
                [
                    self.is_correctly_named,
                    self.exists_in_statusdb,
                    self.files_exist_in_housekeeper,
                    self.files_exist_on_disk,
                ]
            )
            LOG.info(
                f"Flow cell {self.id} has passed all checks, setting flag to True!"
            ) if self._passed_check else LOG.error(
                f"Flow cell {self.id} failed one or more tests, setting flag to {self._passed_check}!"
            )
        return self._passed_check

    def remove_files_from_housekeeper(self) -> None:
        """Remove fastq files and the sample sheet from Housekeeper when deleting a
        flow cell from demultiplexed-runs."""
        if self.fastq_files_exist_in_housekeeper:
            for fastq_file in self.hk_fastq_files:
                LOG.info(f"Removing {fastq_file} from Housekeeper.")
                self.hk.delete_file(fastq_file.id)
            sample_sheets = self.hk.files(tags=[self.id, SequencingFileTag.SAMPLE_SHEET])
            for sample_sheet in sample_sheets:
                LOG.info(f"Removing {sample_sheet} from Housekeeper.")
                self.hk.delete_file(sample_sheet.id)
        self.hk.commit()

    def remove_from_demultiplexed_runs(self) -> None:
        """Removes a flow cell directory completely from demultiplexed-runs."""
        if self.is_correctly_named:
            self.archive_sample_sheet()
        shutil.rmtree(self.path, ignore_errors=True)
        if self.exists_in_statusdb and self.is_correctly_named:
            self.status_db.get_flow_cell(self.id).status = FlowCellStatus.REMOVED
        self.status_db.commit()

    def remove_failed_flow_cell(self) -> None:
        """Performs the two removal actions for failed flow cells."""
        self.remove_from_demultiplexed_runs()
        if self.files_exist_in_housekeeper and self.is_correctly_named:
            self.remove_files_from_housekeeper()

    def archive_sample_sheet(self) -> None:
        """Archives a sample sheet to /home/proj/production/sample_sheets and adds it to
        Housekeeper with an appropriate tag."""
        LOG.info(f"Archiving sample sheet for flow cell {self.run_name}")
        globbed_unaligned_paths: Path.glob = Path(self.path).glob(
            DemultiplexingDirsAndFiles.UNALIGNED_DIR_NAME + ASTERISK
        )
        globbed_unaligned_paths_list: list = list(globbed_unaligned_paths)
        if not globbed_unaligned_paths_list:
            LOG.warning(
                f"No Unaligned directory found for flow cell {self.run_nam}! No sample sheet to archive!"
            )
            return
        unaligned_path: Path = globbed_unaligned_paths_list[0]
        original_sample_sheet: Path = (
            Path(
                unaligned_path,
                DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME,
            )
            if self.sequencer_type == Sequencers.NOVASEQ
            else Path(self.path, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME)
        )

        sample_sheet_run_dir: Path = Path(self.sample_sheets_dir, self.run_name)
        archived_sample_sheet: Path = Path(
            sample_sheet_run_dir, DemultiplexingDirsAndFiles.SAMPLE_SHEET_FILE_NAME
        )

        sample_sheet_run_dir.mkdir(parents=True, exist_ok=True)
        LOG.info(f"Sample sheet directory created for flow cell directory {self.run_name}")

        try:
            shutil.move(original_sample_sheet, archived_sample_sheet)
            self.add_sample_sheet_to_housekeeper(archived_sample_sheet)
            self.hk.commit()
        except FileNotFoundError:
            LOG.warning(f"No sample sheet found in flow cell directory {self.run_name}!")
            shutil.rmtree(sample_sheet_run_dir, ignore_errors=True)

    def add_sample_sheet_to_housekeeper(self, sample_sheet_path: Path):
        """Adds an archive sample sheet to Housekeeper."""

        hk_bundle: hk_models.Bundle = self.hk.bundle(self.id)
        if hk_bundle is None:
            hk_bundle = self.hk.create_new_bundle_and_version(name=self.id)

        with self.hk.session_no_autoflush():
            hk_version: hk_models.Version = self.hk.last_version(bundle=hk_bundle.name)
            if self.hk.files(path=str(sample_sheet_path)).first() is None:
                LOG.info(f"Adding archived sample sheet: {str(sample_sheet_path)}")
                tags: List[str] = [SequencingFileTag.ARCHIVED_SAMPLE_SHEET, self.id]
                self.hk.add_file(path=str(sample_sheet_path), version_obj=hk_version, tags=tags)
        self.hk.commit()



"""Code for uploading to scout via CLI"""
import logging
from pathlib import Path
from typing import Optional

import click
from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.cli.upload.utils import suggest_cases_to_upload
from cg.constants.constants import FileFormat
from cg.exc import CgDataError, ScoutUploadError
from cg.io.controller import WriteStream
from cg.meta.upload.scout.uploadscoutapi import UploadScoutAPI
from cg.models.cg_config import CGConfig
from cg.models.scout.scout_load_config import ScoutLoadConfig
from cg.store import Store
from cg.store.models import Family

LOG = logging.getLogger(__name__)


@click.command()
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@click.option("-p", "--print", "print_console", is_flag=True, help="print config values")
@click.argument("case_id", required=False)
@click.pass_context
def scout(context, re_upload: bool, print_console: bool, case_id: str):
    """Upload variants from analysis to Scout."""
    status_db: Store = context.obj.status_db

    LOG.info("----------------- SCOUT -----------------------")

    if not case_id:
        suggest_cases_to_upload(status_db=status_db)
        return

    context.invoke(
        create_scout_load_config, case_id=case_id, print_console=print_console, re_upload=re_upload
    )
    if not print_console:
        context.invoke(upload_case_to_scout, case_id=case_id, re_upload=re_upload)


@click.command(name="create-scout-load-config")
@click.argument("case-id")
@click.option("-p", "--print", "print_console", is_flag=True, help="Only print config")
@click.option("-r", "--re-upload", is_flag=True, help="Overwrite existing load configs")
@click.pass_obj
def create_scout_load_config(context: CGConfig, case_id: str, print_console: bool, re_upload: bool):
    """Create a load config for a case in scout and add it to housekeeper"""

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    status_db: Store = context.status_db

    LOG.info("----------------- CREATE CONFIG -----------------------")

    LOG.info("Fetching family object")
    case_obj: Family = status_db.family(case_id)
    LOG.info("Create load config")
    if not case_obj.analyses:
        LOG.warning("Could not find analyses for %s", case_id)
        raise click.Abort
    try:
        scout_load_config: ScoutLoadConfig = scout_upload_api.generate_config(case_obj.analyses[0])
    except SyntaxError as error:
        LOG.warning("%s", error)
        raise click.Abort from error
    LOG.info("Found load config %s", scout_load_config)
    root_dir: str = context.meta_apis["upload_api"].analysis_api.root
    LOG.info("Set root dir to %s", root_dir)
    file_path: Path = Path(root_dir, case_id, "scout_load.yaml")

    if print_console:
        click.echo(
            WriteStream.write_stream_from_content(
                content=scout_load_config.dict(exclude_none=True), file_format=FileFormat.YAML
            )
        )
        LOG.info("Would save file to %s", file_path)
        return

    if file_path.exists():
        LOG.warning("Scout load config %s already exists", file_path)
        if re_upload:
            LOG.info("Deleting old load config")
            file_path.unlink()
        else:
            LOG.info(
                "You might remove the file and try again, consider that you might also have it in housekeeper"
            )
            raise click.Abort
    LOG.info("Saving config file to disc")
    scout_upload_api.save_config_file(upload_config=scout_load_config, file_path=file_path)

    try:
        LOG.info("Add config file to housekeeper")
        scout_upload_api.add_scout_config_to_hk(
            config_file_path=file_path, case_id=case_id, delete=re_upload
        )
    except FileExistsError as error:
        LOG.warning("%s, consider removing the file from housekeeper and try again", str(error))
        raise click.Abort


@click.command(name="upload-case-to-scout")
@click.option(
    "-r",
    "--re-upload",
    is_flag=True,
    help="re-upload existing analysis",
)
@click.option("--dry-run", is_flag=True)
@click.argument("case_id")
@click.pass_obj
def upload_case_to_scout(context: CGConfig, re_upload: bool, dry_run: bool, case_id: str):
    """Upload variants and case from analysis to Scout."""

    LOG.info("----------------- UPLOAD -----------------------")

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    scout_api: ScoutAPI = context.scout_api

    tag_name = UploadScoutAPI.get_load_config_tag()
    version_obj = housekeeper_api.last_version(case_id)
    scout_config_file: Optional[hk_models.File] = housekeeper_api.fetch_file_from_version(
        version_obj=version_obj, tags={tag_name}
    )

    if scout_config_file is None:
        raise FileNotFoundError(f"No scout load config was found in housekeeper for {case_id}")

    LOG.info("uploading case %s to scout", case_id)

    if not dry_run:
        scout_api.upload(scout_load_config=scout_config_file.full_path, force=re_upload)

    LOG.info("uploaded to scout using load config %s", scout_config_file.full_path)
    LOG.info("Case loaded successfully to Scout")


@click.command(name="rna-to-scout")
@click.option("--dry-run", is_flag=True)
@click.option("-r", "--research", is_flag=True, help="Upload research report instead of clinical")
@click.option(
    "-u",
    "--update-fusion-report",
    is_flag=True,
    help="re-upload existing fusion report",
)
@click.argument("case_id")
@click.pass_context
def upload_rna_to_scout(
    context, case_id: str, dry_run: bool, update_fusion_report: bool, research: bool
) -> int:
    """Upload an RNA case's gene fusion report and junction splice files for all samples connect via subject_id

    Args:
        case_id                 (string):       RNA case identifier
        dry_run                 (bool):         Skip uploading
        research                (bool):         Upload research report instead of clinical
        update_fusion_report    (bool):         Overwrite existing fusion report
    Returns:

    """

    LOG.info("----------------- UPLOAD RNA TO SCOUT -----------------------")

    result: int = context.invoke(
        upload_rna_fusion_report_to_scout,
        case_id=case_id,
        dry_run=dry_run,
        research=research,
        update=update_fusion_report,
    )
    if result == 0:
        result = context.invoke(upload_rna_junctions_to_scout, case_id=case_id, dry_run=dry_run)
    return result


@click.command(name="rna-fusion-report-to-scout")
@click.option("--dry-run", is_flag=True)
@click.option("--research", is_flag=True)
@click.option(
    "-u",
    "--update",
    is_flag=True,
    help="Overwrite existing report",
)
@click.argument("case_id")
@click.pass_obj
def upload_rna_fusion_report_to_scout(
    context: CGConfig, dry_run: bool, case_id: str, update: bool, research: bool
) -> int:
    """Upload fusion report file for a case to Scout.
    This can also be run as
    `housekeeper get file -V --tag fusion --tag pdf --tag clinical/research <case_id>`
    `scout load gene-fusion-report [-r] <case_id> <path/to/research_gene_fusion_report.pdf>`

    Args:
        dry_run     (bool):         Skip uploading
        case_id     (string):       RNA case identifier
        research    (bool):         Upload research report instead of clinical
        update      (bool):         Overwrite existing report
    Returns:

    """
    LOG.info("----------------- UPLOAD RNA FUSION REPORT TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    try:
        scout_upload_api.upload_fusion_report_to_scout(
            dry_run=dry_run, research=research, case_id=case_id, update=update
        )
    except (CgDataError, ScoutUploadError) as error:
        LOG.error(error)
        return 1
    return 0


@click.command(name="rna-junctions-to-scout")
@click.option("--dry-run", is_flag=True)
@click.argument("case_id")
@click.pass_obj
def upload_rna_junctions_to_scout(context: CGConfig, case_id: str, dry_run: bool) -> int:
    """Upload RNA junctions splice files to Scout.
        This can also be run as
        `housekeeper get file -V --tag junction --tag bed <sample_id>`
        `scout update individual -c <case_id> -n <customer_sample_id> splice_junctions_bed <path/to/junction_file.bed>`
        `housekeeper get file -V --tag coverage --tag bigwig <sample_id>`
        scout update individual -c <case_id> -n <customer_sample_id> rna_coverage_bigwig <path/to/coverage_file.bigWig>
    `   ```

        Args:
            dry_run     (bool):         Skip uploading
            case_id     (string):       RNA case identifier
        Returns:

    """
    LOG.info("----------------- UPLOAD RNA JUNCTIONS TO SCOUT -----------------------")

    scout_upload_api: UploadScoutAPI = context.meta_apis["upload_api"].scout_upload_api
    try:
        scout_upload_api.upload_rna_junctions_to_scout(dry_run=dry_run, case_id=case_id)
    except (CgDataError, ScoutUploadError) as error:
        LOG.error(error)
        return 1
    return 0



""" Parse information fror RunInfo.xml"""

import logging
import xml.etree.ElementTree as ET
from pathlib import Path

LOG = logging.getLogger(__name__)


class RunInfo:
    def __init__(self, runinfo_path: Path):
        self.runinfo_path = runinfo_path
        self.root = self.parse_file()

    def parse_file(self) -> ET.Element:
        """Parse the file RunInfo.xml"""
        LOG.info("Parsing demux conversion stats file %s", self.runinfo_path)
        tree = ET.parse(self.runinfo_path)
        return tree.getroot()

    @property
    def basemask(self) -> str:
        """Return base mask from RunInfo.xml"""
        return ",".join([read.attrib["NumCycles"] for read in self.root.findall("Run/Reads/Read")])

    def mean_read_length(self) -> int:
        """Get the mean read length for this flowcell"""
        read_lengths = [
            int(read.attrib["NumCycles"])
            for read in self.root.findall("Run/Reads/Read")
            if read.attrib["IsIndexedRead"] == "N"
        ]
        return round(sum(read_lengths) / len(read_lengths), 0)



from typing import List

from cgmodels.cg.constants import Pipeline
from pydantic import BaseModel, constr, validator
from pydantic.typing import Optional

from cg.constants import DataDelivery
from cg.constants.constants import GenomeVersion
from cg.models.orders.order import OrderType
from cg.models.orders.sample_base import (
    ContainerEnum,
    NAME_PATTERN,
    PriorityEnum,
    SexEnum,
    StatusEnum,
    ControlEnum,
)
from cg.store import models


class OptionalIntValidator:
    @classmethod
    def str_to_int(cls, v: str) -> Optional[int]:
        return int(v) if v else None


class OptionalFloatValidator:
    @classmethod
    def str_to_float(cls, v: str) -> Optional[float]:
        return float(v) if v else None


class OrderInSample(BaseModel):
    # Order portal specific
    internal_id: Optional[
        constr(max_length=models.Sample.internal_id.property.columns[0].type.length)
    ]
    _suitable_project: OrderType = None
    application: constr(max_length=models.Application.tag.property.columns[0].type.length)
    comment: Optional[constr(max_length=models.Sample.comment.property.columns[0].type.length)]
    data_analysis: Pipeline
    data_delivery: DataDelivery
    name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=models.Sample.name.property.columns[0].type.length,
    )
    priority: PriorityEnum = PriorityEnum.standard
    require_qcok: bool = False
    volume: str

    @classmethod
    def is_sample_for(cls, project: OrderType):
        return project == cls._suitable_project


class Of1508Sample(OrderInSample):
    # Orderform 1508
    # Order portal specific
    internal_id: Optional[
        constr(max_length=models.Sample.internal_id.property.columns[0].type.length)
    ]
    # "required for new samples"
    name: Optional[
        constr(
            regex=NAME_PATTERN,
            min_length=2,
            max_length=models.Sample.name.property.columns[0].type.length,
        )
    ]

    # customer
    age_at_sampling: Optional[float]
    # "application": str,
    family_name: constr(
        regex=NAME_PATTERN,
        min_length=2,
        max_length=models.Family.name.property.columns[0].type.length,
    )
    case_internal_id: Optional[
        constr(max_length=models.Sample.internal_id.property.columns[0].type.length)
    ]
    sex: SexEnum = SexEnum.unknown
    tumour: bool = False
    source: Optional[str]
    control: Optional[str]
    volume: Optional[str]
    container: Optional[ContainerEnum]
    # "required if plate for new samples"
    container_name: Optional[str]
    well_position: Optional[str]
    # "Required if samples are part of trio/family"
    mother: Optional[
        constr(regex=NAME_PATTERN, max_length=models.Sample.name.property.columns[0].type.length)
    ]
    father: Optional[
        constr(regex=NAME_PATTERN, max_length=models.Sample.name.property.columns[0].type.length)
    ]
    # This information is required for panel analysis
    capture_kit: Optional[str]
    # This information is required for panel- or exome analysis
    elution_buffer: Optional[str]
    tumour_purity: Optional[int]
    # "This information is optional for FFPE-samples for new samples"
    formalin_fixation_time: Optional[int]
    post_formalin_fixation_time: Optional[int]
    tissue_block_size: Optional[str]
    # "Not Required"
    cohorts: Optional[List[str]]
    phenotype_groups: Optional[List[str]]
    phenotype_terms: Optional[List[str]]
    require_qcok: bool = False
    quantity: Optional[int]
    subject_id: Optional[
        constr(
            regex=NAME_PATTERN, max_length=models.Sample.subject_id.property.columns[0].type.length
        )
    ]
    synopsis: Optional[str]

    @validator("container", "container_name", "name", "source", "subject_id", "volume")
    def required_for_new_samples(cls, value, values, **kwargs):
        if not value and not values.get("internal_id"):
            raise ValueError("required for new sample '%s'" % (values.get("name")))
        return value

    @validator(
        "tumour_purity",
        "formalin_fixation_time",
        "post_formalin_fixation_time",
        "quantity",
        pre=True,
    )
    def str_to_int(cls, v: str) -> Optional[int]:
        return OptionalIntValidator.str_to_int(v=v)

    @validator(
        "age_at_sampling",
        "volume",
        pre=True,
    )
    def str_to_float(cls, v: str) -> Optional[float]:
        return OptionalFloatValidator.str_to_float(v=v)


class MipDnaSample(Of1508Sample):
    _suitable_project = OrderType.MIP_DNA
    # "Required if data analysis in Scout or vcf delivery"
    panels: List[
        constr(min_length=1, max_length=models.Panel.abbrev.property.columns[0].type.length)
    ]
    status: StatusEnum


class BalsamicSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC


class BalsamicQCSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC_QC
    reference_genome: Optional[GenomeVersion]


class BalsamicUmiSample(Of1508Sample):
    _suitable_project = OrderType.BALSAMIC_UMI


class MipRnaSample(Of1508Sample):
    _suitable_project = OrderType.MIP_RNA


class FastqSample(OrderInSample):
    _suitable_project = OrderType.FASTQ

    # Orderform 1508
    # "required"
    container: Optional[ContainerEnum]
    sex: SexEnum = SexEnum.unknown
    source: str
    tumour: bool
    # "required if plate"
    container_name: Optional[str]
    well_position: Optional[str]
    elution_buffer: str
    # This information is required for panel analysis
    capture_kit: Optional[str]
    # "Not Required"
    quantity: Optional[int]

    @validator("quantity", pre=True)
    def str_to_int(cls, v: str) -> Optional[int]:
        return OptionalIntValidator.str_to_int(v=v)


class RmlSample(OrderInSample):
    _suitable_project = OrderType.RML

    # 1604 Orderform Ready made libraries (RML)
    # Order portal specific
    # "This information is required"
    pool: constr(max_length=models.Pool.name.property.columns[0].type.length)
    concentration: float
    concentration_sample: Optional[float]
    index: str
    index_number: Optional[str]
    # "Required if Plate"
    rml_plate_name: Optional[str]
    well_position_rml: Optional[str]
    # "Automatically generated (if not custom) or custom"
    index_sequence: Optional[str]
    # "Not required"
    control: Optional[str]

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> Optional[float]:
        return OptionalFloatValidator.str_to_float(v=v)


class FluffySample(RmlSample):
    _suitable_project = OrderType.FLUFFY
    # 1604 Orderform Ready made libraries (RML)


class MetagenomeSample(OrderInSample):
    _suitable_project = OrderType.METAGENOME

    # 1605 Orderform Microbial Metagenomes- 16S
    # "This information is required"
    container: Optional[ContainerEnum]
    elution_buffer: str
    source: str
    # "Required if Plate"
    container_name: Optional[str]
    well_position: Optional[str]
    # "This information is not required"
    concentration_sample: Optional[float]
    quantity: Optional[int]
    extraction_method: Optional[str]
    control: Optional[ControlEnum]

    @validator("quantity", pre=True)
    def str_to_int(cls, v: str) -> Optional[int]:
        return OptionalIntValidator.str_to_int(v=v)

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> Optional[float]:
        return OptionalFloatValidator.str_to_float(v=v)


class MicrobialSample(OrderInSample):
    # 1603 Orderform Microbial WGS
    # "These fields are required"
    organism: constr(max_length=models.Organism.internal_id.property.columns[0].type.length)
    reference_genome: constr(
        max_length=models.Sample.reference_genome.property.columns[0].type.length
    )
    elution_buffer: str
    extraction_method: Optional[str]
    container: ContainerEnum
    # "Required if Plate"
    container_name: Optional[str]
    well_position: Optional[str]
    # "Required if "Other" is chosen in column "Species""
    organism_other: Optional[
        constr(max_length=models.Organism.internal_id.property.columns[0].type.length)
    ]
    # "These fields are not required"
    concentration_sample: Optional[float]
    quantity: Optional[int]
    verified_organism: Optional[bool]  # sent to LIMS
    control: Optional[str]

    @validator("quantity", pre=True)
    def str_to_int(cls, v: str) -> Optional[int]:
        return OptionalIntValidator.str_to_int(v=v)

    @validator("concentration_sample", pre=True)
    def str_to_float(cls, v: str) -> Optional[float]:
        return OptionalFloatValidator.str_to_float(v=v)


class MicrosaltSample(MicrobialSample):
    _suitable_project = OrderType.MICROSALT
    # 1603 Orderform Microbial WGS


class SarsCov2Sample(MicrobialSample):
    _suitable_project = OrderType.SARS_COV_2

    # 2184 Orderform SARS-COV-2
    # "These fields are required"
    collection_date: str
    lab_code: str
    primer: str
    original_lab: str
    original_lab_address: str
    pre_processing_method: str
    region: str
    region_code: str
    selection_criteria: str
    volume: Optional[str]


def sample_class_for(project: OrderType):
    """Get the sample class for the specified project

    Args:
        project     (OrderType):    Project to get sample subclass for
    Returns:
        Subclass of OrderInSample
    """

    def all_subclasses(cls):
        """Get all subclasses recursively for a class

        Args:
            cls     (Class):    Class to get all subclasses for
        Returns:
            Set of Subclasses of cls
        """
        if cls.__subclasses__():
            return set(cls.__subclasses__()).union(
                [s for c in cls.__subclasses__() for s in all_subclasses(c)]
            )

        return []

    for sub_cls in all_subclasses(OrderInSample):
        if sub_cls.is_sample_for(project):
            return sub_cls

    raise ValueError



"""API to run Vogue"""

import datetime as dt
from typing import Optional

from cg.apps.gt import GenotypeAPI
from cg.apps.vogue import VogueAPI
from cg.constants.constants import FileFormat
from cg.io.controller import ReadStream
from cg.store import Store, models


class UploadVogueAPI:
    """API to load data into Vogue"""

    def __init__(self, genotype_api: GenotypeAPI, vogue_api: VogueAPI, store: Store):
        self.genotype_api = genotype_api
        self.vogue_api = vogue_api
        self.store = store

    def load_genotype(self, days: int) -> None:
        """Loading genotype data from the genotype database into the trending database"""

        samples: dict = ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.genotype_api.export_sample(days=days)
        )
        for sample_id, sample_dict in samples.items():
            sample_dict["_id"] = sample_id
            self.vogue_api.load_genotype_data(sample_dict)

        samples_analysis: dict = ReadStream.get_content_from_stream(
            file_format=FileFormat.JSON, stream=self.genotype_api.export_sample_analysis(days=days)
        )
        for sample_id, sample_dict in samples_analysis.items():
            sample_dict["_id"] = sample_id
            self.vogue_api.load_genotype_data(sample_dict)

    def load_apptags(self) -> None:
        """Loading application tags from statusdb into the trending database"""
        apptags = self.store.applications()
        apptags_for_vogue = [
            {"tag": tag.tag, "prep_category": tag.prep_category} for tag in apptags.all()
        ]

        self.vogue_api.load_apptags(apptags_for_vogue)

    def load_bioinfo_raw(self, load_bioinfo_inputs):
        """Running vogue load bioinfo raw."""

        self.vogue_api.load_bioinfo_raw(load_bioinfo_inputs)

    def load_bioinfo_process(self, load_bioinfo_inputs, cleanup):
        """Running vogue load bioinfo process."""

        self.vogue_api.load_bioinfo_process(load_bioinfo_inputs, cleanup)

    def load_bioinfo_sample(self, load_bioinfo_inputs):
        """Running vogue load bioinfo sample."""

        self.vogue_api.load_bioinfo_sample(load_bioinfo_inputs)

    @staticmethod
    def update_analysis_uploaded_to_vogue_date(
        analysis: models.Analysis,
        vogue_upload_date: Optional[dt.datetime] = dt.datetime.now(),
    ) -> models.Analysis:
        analysis.uploaded_to_vogue_at = vogue_upload_date
        return analysis



from pathlib import Path

from cg.io.json import read_json, write_json, read_json_stream, write_json_stream


def test_get_content_from_file(json_file_path: Path):
    """
    Tests read_json
    """
    # GIVEN a json file

    # WHEN reading the json file
    raw_json_content: dict = read_json(file_path=json_file_path)

    # Then assert a dict is returned
    assert isinstance(raw_json_content, dict)


def test_get_content_from_stream(json_stream: str):
    """
    Tests read_json_stream
    """
    # GIVEN a string in json format

    # WHEN reading the json content in string
    raw_content: dict = read_json_stream(stream=json_stream)

    # THEN assert a dict is returned
    assert isinstance(raw_content, dict)


def test_write_json(json_file_path: Path, json_temp_path: Path):
    """
    Tests write_json
    """
    # GIVEN a json file

    # GIVEN a file path to write to

    # WHEN reading the json file
    raw_json_content: dict = read_json(file_path=json_file_path)

    # WHEN writing the json file from dict
    write_json(content=raw_json_content, file_path=json_temp_path)

    # THEN assert that a file was successfully created
    assert Path.exists(json_temp_path)

    # WHEN reading it as a json
    written_raw_json_content: dict = read_json(file_path=json_temp_path)

    # THEN assert that all data is kept
    assert raw_json_content == written_raw_json_content


def test_write_json_stream(json_stream: str):
    """
    Tests write_json_stream
    """
    # GIVEN a dict
    raw_content: dict = read_json_stream(stream=json_stream)

    # WHEN writing the dict to a json stream
    json_content = write_json_stream(content=raw_content)

    # THEN assert that all data is kept and properly formatted
    assert json_stream == json_content



from typing import List

import pydantic
import pytest

from cg.models.rnafusion.rnafusion_sample import RnafusionSample


def test_instantiate_rnafusion_sample(
    rnafusion_sample: str,
    rnafusion_fastq_r1: List[str],
    rnafusion_fastq_r2_same_length: List[str],
    rnafusion_strandedness_acceptable: str,
):
    """
    Tests rnafusion sample.
    """
    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a MipAnalysis object
    rnafusion_sample_object = RnafusionSample(
        sample=rnafusion_sample,
        fastq_r1=rnafusion_fastq_r1,
        fastq_r2=rnafusion_fastq_r2_same_length,
        strandedness=rnafusion_strandedness_acceptable,
    )

    # THEN assert that it was successfully created
    assert isinstance(rnafusion_sample_object, RnafusionSample)


def test_instantiate_rnafusion_sample_fastq_r1_r2_different_length(
    rnafusion_sample: str,
    rnafusion_fastq_r1: List[str],
    rnafusion_fastq_r2_not_same_length: List[str],
    rnafusion_strandedness_acceptable: str,
):
    """
    Tests rnafusion sample with different fastq_r1 and fastq_r2 length.
    """

    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a sample object THEN throws a ValidationError

    with pytest.raises(pydantic.ValidationError):
        RnafusionSample(
            sample=rnafusion_sample,
            fastq_r1=rnafusion_fastq_r1,
            fastq_r2=rnafusion_fastq_r2_not_same_length,
            strandedness=rnafusion_strandedness_acceptable,
        )


def test_instantiate_rnafusion_sample_fastq_r2_empty(
    rnafusion_sample: str,
    rnafusion_fastq_r1: List[str],
    rnafusion_fastq_r2_empty: list,
    rnafusion_strandedness_acceptable: str,
):
    """
    Tests rnafusion sample with fastq_r2 empty (single end).
    """

    # GIVEN a sample with fastq files and strandedness

    # WHEN instantiating a sample object
    rnafusion_sample_object: RnafusionSample = RnafusionSample(
        sample=rnafusion_sample,
        fastq_r1=rnafusion_fastq_r1,
        fastq_r2=rnafusion_fastq_r2_empty,
        strandedness=rnafusion_strandedness_acceptable,
    )

    # THEN assert that it was successfully created
    assert isinstance(rnafusion_sample_object, RnafusionSample)


def test_instantiate_rnafusion_strandedness_not_acceptable(
    rnafusion_sample: str,
    rnafusion_fastq_r1: List[str],
    rnafusion_fastq_r2_same_length: List[str],
    rnafusion_strandedness_not_acceptable: str,
):
    """
    Tests rnafusion sample with unacceptable strandedness.
    """
    # WHEN instantiating a sample object THEN throws a ValidationError

    with pytest.raises(pydantic.ValidationError):
        RnafusionSample(
            sample=rnafusion_sample,
            fastq_r1=rnafusion_fastq_r1,
            fastq_r2=rnafusion_fastq_r2_same_length,
            strandedness=rnafusion_strandedness_not_acceptable,
        )



"""Utility functions to simply add test data in a cg store."""
import logging
from datetime import datetime
from typing import List, Optional

from housekeeper.store import models as hk_models

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import DataDelivery, Pipeline
from cg.constants.pedigree import Pedigree
from cg.constants.priority import PriorityTerms
from cg.constants.sequencing import Sequencers
from cg.constants.subject import Gender, PhenotypeStatus
from cg.store import Store, models
from cg.store.models import Flowcell, Bed, BedVersion, Sample

LOG = logging.getLogger(__name__)


class StoreHelpers:
    """Class to hold helper functions that needs to be used all over."""

    @staticmethod
    def ensure_hk_bundle(
        store: HousekeeperAPI, bundle_data: dict, include: bool = False
    ) -> hk_models.Bundle:
        """Utility function to add a bundle of information to a housekeeper api."""

        bundle_exists = False
        for bundle in store.bundles():
            if bundle.name != bundle_data["name"]:
                continue
            bundle_exists = True
            _bundle = bundle
            break

        if not bundle_exists:
            _bundle, _version = store.add_bundle(bundle_data)
            store.add_commit(_bundle, _version)

        if include:
            store.include(_version)

        return _bundle

    @staticmethod
    def ensure_hk_version(store: HousekeeperAPI, bundle_data: dict) -> hk_models.Version:
        """Utility function to return existing or create an version for tests."""
        _bundle = StoreHelpers.ensure_hk_bundle(store, bundle_data)
        return store.last_version(_bundle.name)

    @staticmethod
    def ensure_application_version(
        store: Store,
        application_tag: str = "dummy_tag",
        application_type: str = "wgs",
        is_external: bool = False,
        is_rna: bool = False,
        description: str = None,
        sequencing_depth: int = None,
        is_accredited: bool = False,
    ) -> models.ApplicationVersion:
        """Utility function to return existing or create application version for tests."""
        if is_rna:
            application_tag = "rna_tag"
            application_type = "wts"

        application = store.application(tag=application_tag)
        if not application:
            application = StoreHelpers.add_application(
                store,
                application_tag,
                application_type,
                is_external=is_external,
                description=description,
                is_accredited=is_accredited,
                sequencing_depth=sequencing_depth,
            )

        prices = {
            PriorityTerms.STANDARD: 10,
            PriorityTerms.PRIORITY: 20,
            PriorityTerms.EXPRESS: 30,
            PriorityTerms.RESEARCH: 5,
        }
        version = store.application_version(application, 1)
        if not version:
            version = store.add_version(application, 1, valid_from=datetime.now(), prices=prices)

            store.add_commit(version)
        return version

    @staticmethod
    def ensure_application(
        store: Store,
        tag: str,
        application_type: str = "wgs",
        description: str = "dummy_description",
        is_archived: bool = False,
    ) -> models.Application:
        """Ensure that application exists in store."""
        application: models.Application = store.application(tag=tag)
        if not application:
            application: models.Application = StoreHelpers.add_application(
                store=store,
                application_tag=tag,
                application_type=application_type,
                description=description,
                is_archived=is_archived,
            )
        return application

    @staticmethod
    def add_application(
        store: Store,
        application_tag: str = "dummy_tag",
        application_type: str = "wgs",
        description: str = None,
        is_archived: bool = False,
        is_accredited: bool = False,
        is_external: bool = False,
        min_sequencing_depth: int = 30,
        **kwargs,
    ) -> models.Application:
        """Utility function to add a application to a store."""
        application = store.application(tag=application_tag)
        if application:
            return application

        if not description:
            description = "dummy_description"
        application = store.add_application(
            tag=application_tag,
            category=application_type,
            description=description,
            is_archived=is_archived,
            percent_kth=80,
            percent_reads_guaranteed=75,
            is_accredited=is_accredited,
            limitations="A limitation",
            is_external=is_external,
            min_sequencing_depth=min_sequencing_depth,
            **kwargs,
        )
        store.add_commit(application)
        return application

    @staticmethod
    def ensure_bed_version(store: Store, bed_name: str = "dummy_bed") -> BedVersion:
        """Return existing or create and return bed version for tests."""
        bed: Optional[Bed] = store.get_bed_by_name(bed_name=bed_name)
        if not bed:
            bed: Bed = store.add_bed(name=bed_name)
            store.add_commit(bed)

        bed_version: Optional[BedVersion] = store.get_latest_bed_version(bed_name=bed_name)
        if not bed_version:
            bed_version: BedVersion = store.add_bed_version(
                bed=bed, version=1, filename="dummy_filename", shortname=bed_name
            )
            store.add_commit(bed_version)
        return bed_version

    @staticmethod
    def ensure_collaboration(store: Store, collaboration_id: str = "all_customers"):
        collaboration = store.collaboration(collaboration_id)
        if not collaboration:
            collaboration = store.add_collaboration(collaboration_id, collaboration_id)
        return collaboration

    @staticmethod
    def ensure_customer(
        store: Store,
        customer_id: str = "cust000",
        name: str = "Production",
        scout_access: bool = False,
    ) -> models.Customer:
        """Utility function to return existing or create customer for tests."""
        collaboration: models.Collaboration = StoreHelpers.ensure_collaboration(store)
        customer = store.customer(customer_id)

        if not customer:
            customer = store.add_customer(
                internal_id=customer_id,
                name=name,
                scout_access=scout_access,
                invoice_address="Test street",
                invoice_reference="ABCDEF",
            )
            customer.collaborations.append(collaboration)
            store.add_commit(customer)
        return customer

    @staticmethod
    def add_analysis(
        store: Store,
        case: models.Family = None,
        started_at: datetime = None,
        completed_at: datetime = None,
        uploaded_at: datetime = None,
        upload_started: datetime = None,
        delivery_reported_at: datetime = None,
        cleaned_at: datetime = None,
        pipeline: Pipeline = Pipeline.BALSAMIC,
        pipeline_version: str = "1.0",
        data_delivery: DataDelivery = DataDelivery.FASTQ_QC,
        uploading: bool = False,
        config_path: str = None,
    ) -> models.Analysis:
        """Utility function to add an analysis for tests."""

        if not case:
            case = StoreHelpers.add_case(store, data_analysis=pipeline, data_delivery=data_delivery)

        analysis = store.add_analysis(pipeline=pipeline, version=pipeline_version)

        analysis.started_at = started_at or datetime.now()
        if completed_at:
            analysis.completed_at = completed_at
        if uploaded_at:
            analysis.uploaded_at = uploaded_at
        if delivery_reported_at:
            analysis.delivery_report_created_at = delivery_reported_at
        if cleaned_at:
            analysis.cleaned_at = cleaned_at
        if uploading:
            analysis.upload_started_at = upload_started or datetime.now()
        if config_path:
            analysis.config_path = config_path
        if pipeline:
            analysis.pipeline = str(pipeline)

        analysis.limitations = "A limitation"
        analysis.family = case
        store.add_commit(analysis)
        return analysis

    @staticmethod
    def add_sample(
        store: Store,
        application_tag: str = "dummy_tag",
        application_type: str = "tgs",
        control: str = "",
        customer_id: str = None,
        gender: str = Gender.FEMALE,
        is_external: bool = False,
        is_rna: bool = False,
        is_tumour: bool = False,
        internal_id: str = None,
        reads: int = None,
        name: str = "sample_test",
        original_ticket: str = None,
        **kwargs,
    ) -> models.Sample:
        """Utility function to add a sample to use in tests."""
        customer_id = customer_id or "cust000"
        customer = StoreHelpers.ensure_customer(store, customer_id=customer_id)
        application_version = StoreHelpers.ensure_application_version(
            store=store,
            application_tag=application_tag,
            application_type=application_type,
            is_external=is_external,
            is_rna=is_rna,
        )
        application_version_id = application_version.id

        if internal_id:
            existing_sample: models.Sample = store.sample(internal_id=internal_id)
            if existing_sample:
                return existing_sample

        sample = store.add_sample(
            name=name,
            sex=gender,
            control=control,
            original_ticket=original_ticket,
            tumour=is_tumour,
            reads=reads,
            internal_id=internal_id,
        )

        sample.application_version_id = application_version_id
        sample.customer = customer
        sample.ordered_at = datetime.now()

        for key, value in kwargs.items():
            if key == "flowcell":
                sample.flowcells.append(kwargs["flowcell"])
            elif hasattr(sample, key):
                setattr(sample, key, value)
            else:
                raise AttributeError(f"Unknown sample attribute/feature: {key}, {value}")

        store.add_commit(sample)
        return sample

    @staticmethod
    def ensure_panel(
        store: Store, panel_id: str = "panel_test", customer_id: str = "cust000"
    ) -> models.Panel:
        """Utility function to add a panel to use in tests."""
        customer = StoreHelpers.ensure_customer(store, customer_id)
        panel = store.panel(panel_id)
        if not panel:
            panel = store.add_panel(
                customer=customer,
                name=panel_id,
                abbrev=panel_id,
                version=1.0,
                date=datetime.now(),
                genes=1,
            )
            store.add_commit(panel)
        return panel

    @staticmethod
    def add_case(
        store: Store,
        name: str = "case_test",
        data_analysis: str = Pipeline.MIP_DNA,
        data_delivery: DataDelivery = DataDelivery.SCOUT,
        action: str = None,
        internal_id: str = None,
        customer_id: str = "cust000",
        panels: List[str] = [],
        case_obj: models.Family = None,
        ticket: str = "123456",
    ) -> models.Family:
        """Utility function to add a case to use in tests,
        If no case object is used a autogenerated case id will be used.

        """
        if not panels:
            panels: List[str] = ["panel_test"]
        customer = StoreHelpers.ensure_customer(store, customer_id)
        if case_obj:
            panels = case_obj.panels
        for panel_name in panels:
            StoreHelpers.ensure_panel(store=store, panel_id=panel_name, customer_id=customer_id)

        if not case_obj:
            case_obj: Optional[models.Family] = store.family(internal_id=name)
        if not case_obj:
            case_obj = store.add_case(
                data_analysis=data_analysis,
                data_delivery=data_delivery,
                name=name,
                panels=panels,
                ticket=ticket,
            )
        if action:
            case_obj.action = action
        if internal_id:
            case_obj.internal_id = internal_id

        case_obj.customer = customer
        store.add_commit(case_obj)
        return case_obj

    @staticmethod
    def ensure_case(
        store: Store,
        name: str = "test-case",
        case_id: str = "blueeagle",
        customer: models.Customer = None,
        data_analysis: Pipeline = Pipeline.MIP_DNA,
        data_delivery: DataDelivery = DataDelivery.SCOUT,
    ):
        """Load a case with samples and link relations."""
        if not customer:
            customer = StoreHelpers.ensure_customer(store=store)
        case = store.family(internal_id=case_id) or store.find_family(customer=customer, name=name)
        if not case:
            case = StoreHelpers.add_case(
                store=store,
                data_analysis=data_analysis,
                data_delivery=data_delivery,
                name=name,
                internal_id=case_id,
            )
            case.customer = customer
        return case

    @staticmethod
    def ensure_case_from_dict(
        store: Store,
        case_info: dict,
        app_tag: str = None,
        ordered_at: datetime = None,
        completed_at: datetime = None,
        created_at: datetime = datetime.now(),
    ):
        """Load a case with samples and link relations from a dictionary."""
        customer_obj = StoreHelpers.ensure_customer(store)
        case_obj = store.Family(
            name=case_info["name"],
            panels=case_info["panels"],
            internal_id=case_info["internal_id"],
            ordered_at=ordered_at,
            data_analysis=case_info.get("data_analysis", str(Pipeline.MIP_DNA)),
            data_delivery=case_info.get("data_delivery", str(DataDelivery.SCOUT)),
            created_at=created_at,
            action=case_info.get("action"),
            tickets=case_info["tickets"],
        )

        case_obj = StoreHelpers.add_case(
            store, case_obj=case_obj, customer_id=customer_obj.internal_id
        )

        app_tag = app_tag or "WGSPCFC030"
        app_type = case_info.get("application_type", "wgs")
        StoreHelpers.ensure_application_version(store, application_tag=app_tag)

        sample_objs = {}
        for sample_data in case_info["samples"]:
            sample_id = sample_data["internal_id"]
            sample_obj = StoreHelpers.add_sample(
                store,
                gender=sample_data["sex"],
                name=sample_data.get("name"),
                internal_id=sample_id,
                application_type=app_type,
                application_tag=app_tag,
                original_ticket=sample_data["original_ticket"],
                reads=sample_data["reads"],
                capture_kit=sample_data["capture_kit"],
            )
            sample_objs[sample_id] = sample_obj

        for sample_data in case_info["samples"]:
            sample_obj = sample_objs[sample_data["internal_id"]]
            father = None
            if sample_data.get(Pedigree.FATHER):
                father = sample_objs[sample_data[Pedigree.FATHER]]
            mother = None
            if sample_data.get(Pedigree.MOTHER):
                mother = sample_objs[sample_data[Pedigree.MOTHER]]
            StoreHelpers.add_relationship(
                store,
                case=case_obj,
                sample=sample_obj,
                status=sample_data.get("status", PhenotypeStatus.UNKNOWN),
                father=father,
                mother=mother,
            )

        StoreHelpers.add_analysis(
            store,
            pipeline=Pipeline.MIP_DNA,
            case=case_obj,
            completed_at=completed_at or datetime.now(),
        )
        return case_obj

    @staticmethod
    def add_organism(
        store: Store,
        internal_id: str = "organism_test",
        name: str = "organism_name",
        reference_genome: str = "reference_genome_test",
    ) -> models.Organism:
        """Utility function to add an organism to use in tests."""
        return store.add_organism(
            internal_id=internal_id, name=name, reference_genome=reference_genome
        )

    @staticmethod
    def ensure_organism(
        store: Store,
        organism_id: str = "organism_test",
        name: str = "organism_name",
        reference_genome: str = "reference_genome_test",
    ) -> models.Organism:
        """Utility function to add an organism to use in tests."""
        organism = StoreHelpers.add_organism(
            store, internal_id=organism_id, name=name, reference_genome=reference_genome
        )
        store.add_commit(organism)
        return organism

    @staticmethod
    def add_microbial_sample(
        store: Store,
        sample_id: str = "microbial_sample_id",
        priority: str = PriorityTerms.RESEARCH,
        name: str = "microbial_name_test",
        organism: models.Organism = None,
        comment: str = "comment",
        ticket: int = 123456,
    ) -> models.Sample:
        """Utility function to add a sample to use in tests."""
        customer = StoreHelpers.ensure_customer(store, "cust_test")
        application_version = StoreHelpers.ensure_application_version(store)
        if not organism:
            organism = StoreHelpers.ensure_organism(store)

        sample = store.add_sample(
            name=name,
            comment=comment,
            internal_id=sample_id,
            priority=priority,
            application_version=application_version,
            organism=organism,
            reads=6000000,
            sex=Gender.UNKNOWN,
        )
        sample.customer = customer
        case = StoreHelpers.ensure_case(
            store=store,
            name=str(ticket),
            customer=customer,
            data_analysis=Pipeline.MICROSALT,
            data_delivery=DataDelivery.FASTQ_QC,
        )
        StoreHelpers.add_relationship(store=store, case=case, sample=sample)
        return sample

    @staticmethod
    def add_samples(store: Store, nr_samples: int = 5) -> List[models.Sample]:
        """Utility function to add a number of samples to use in tests."""
        nr_samples = max(nr_samples, 2)
        return [
            StoreHelpers.add_sample(store, str(sample_id)) for sample_id in range(1, nr_samples)
        ]

    @staticmethod
    def add_flowcell(
        store: Store,
        flow_cell_id: str = "flowcell_test",
        archived_at: datetime = None,
        sequencer_type: str = Sequencers.HISEQX,
        samples: List[Sample] = None,
        status: str = None,
        date: datetime = datetime.now(),
    ) -> Flowcell:
        """Utility function to add a flow cell to the store and return an object."""
        flow_cell = store.add_flow_cell(
            flow_cell_id=flow_cell_id,
            sequencer_name="dummy_sequencer",
            sequencer_type=sequencer_type,
            date=date,
        )
        flow_cell.archived_at = archived_at
        if samples:
            flow_cell.samples = samples
        if status:
            flow_cell.status = status

        store.add_commit(flow_cell)
        return flow_cell

    @staticmethod
    def add_relationship(
        store: Store,
        sample: models.Sample,
        case: models.Family,
        status: str = PhenotypeStatus.UNKNOWN,
        father: models.Sample = None,
        mother: models.Sample = None,
    ) -> models.FamilySample:
        """Utility function to link a sample to a case."""
        link = store.relate_sample(
            sample=sample, family=case, status=status, father=father, mother=mother
        )
        store.add_commit(link)
        return link

    @staticmethod
    def add_synopsis_to_case(
        store: Store, case_id: str, synopsis: str = "a synopsis"
    ) -> Optional[models.Family]:
        """Function for adding a synopsis to a case in the database."""
        case_obj: models.Family = store.family(internal_id=case_id)
        if not case_obj:
            LOG.warning("Could not find case")
            return None
        case_obj.synopsis = synopsis
        store.commit()
        return case_obj

    @staticmethod
    def add_phenotype_groups_to_sample(
        store: Store, sample_id: str, phenotype_groups: [str] = None
    ) -> Optional[models.Sample]:
        """Function for adding a phenotype group to a sample in the database."""
        if phenotype_groups is None:
            phenotype_groups = ["a phenotype group"]
        sample_obj: models.Sample = store.sample(internal_id=sample_id)
        if not sample_obj:
            LOG.warning("Could not find sample")
            return None
        sample_obj.phenotype_groups = phenotype_groups
        store.commit()
        return sample_obj

    @staticmethod
    def add_phenotype_terms_to_sample(
        store: Store, sample_id: str, phenotype_terms: List[str] = []
    ) -> Optional[models.Sample]:
        """Function for adding a phenotype term to a sample in the database."""
        if not phenotype_terms:
            phenotype_terms: List[str] = ["a phenotype term"]
        sample_obj: models.Sample = store.sample(internal_id=sample_id)
        if not sample_obj:
            LOG.warning("Could not find sample")
            return None
        sample_obj.phenotype_terms = phenotype_terms
        store.commit()
        return sample_obj

    @staticmethod
    def add_subject_id_to_sample(
        store: Store, sample_id: str, subject_id: str = "a subject_id"
    ) -> Optional[models.Sample]:
        """Function for adding a subject_id to a sample in the database."""
        sample_obj: models.Sample = store.sample(internal_id=sample_id)
        if not sample_obj:
            LOG.warning("Could not find sample")
            return None
        sample_obj.subject_id = subject_id
        store.commit()
        return sample_obj

    @classmethod
    def relate_samples(cls, base_store: Store, case: models.Family, samples: List[models.Sample]):
        """Utility function to relate many samples to one case."""

        for sample in samples:
            link = base_store.relate_sample(
                family=case, sample=sample, status=PhenotypeStatus.UNKNOWN
            )
            base_store.add_commit(link)

    @classmethod
    def add_case_with_samples(
        cls,
        base_store: Store,
        case_id: str,
        nr_samples: int,
        sequenced_at: datetime = datetime.now(),
    ) -> models.Family:
        """Utility function to add one case with many samples and return the case."""

        samples: List[models.Sample] = cls.add_samples(store=base_store, nr_samples=nr_samples)
        for sample in samples:
            sample.sequenced_at: datetime = sequenced_at
        case: models.Family = cls.add_case(store=base_store, internal_id=case_id, name=case_id)
        cls.relate_samples(base_store=base_store, case=case, samples=samples)
        return case

    @classmethod
    def add_cases_with_samples(
        cls, base_store: Store, nr_cases: int, sequenced_at: datetime
    ) -> List[models.Family]:
        """Utility function to add many cases with two samples to use in tests."""

        cases: List[models.Family] = []
        for i in range(nr_cases):
            case: List[models.Family] = cls.add_case_with_samples(
                base_store, f"f{i}", 2, sequenced_at=sequenced_at
            )
            cases.append(case)
        return cases

    @classmethod
    def ensure_pool(
        cls,
        store: Store,
        customer_id: str = "cust000",
        name: str = "test_pool",
        ticket: str = "987654",
        application_tag: str = "dummy_tag",
        application_type: str = "tgs",
        is_external: bool = False,
        is_rna: bool = False,
    ) -> models.Pool:
        """Utility function to add a pool that can be used in tests."""
        customer_id = customer_id or "cust000"
        customer = store.customer(customer_id)
        if not customer:
            customer = StoreHelpers.ensure_customer(store, customer_id=customer_id)

        application_version = StoreHelpers.ensure_application_version(
            store=store,
            application_tag=application_tag,
            application_type=application_type,
            is_external=is_external,
            is_rna=is_rna,
        )

        pool = store.add_pool(
            name=name,
            ordered=datetime.now(),
            application_version=application_version,
            customer=customer,
            order="test_order",
        )
        store.add_commit(pool)
        return pool

    @classmethod
    def ensure_user(
        cls,
        store: Store,
        customer: models.Customer,
        email: str = "Bob@bobmail.com",
        name: str = "Bob",
        is_admin: bool = False,
    ) -> models.User:
        """Utility function to add a user that can be used in tests."""
        user = store.user(email=email)
        if not user:
            user = store.add_user(customer=customer, email=email, name=name, is_admin=is_admin)
            store.add_commit(user)
        return user

    @classmethod
    def ensure_invoice(
        cls,
        store: Store,
        invoice_id: int = 0,
        customer_id: str = "cust000",
        discount: int = 0,
        record_type: str = "Sample",
        internal_id: str = "savedkitten",
    ) -> models.Invoice:
        """Utility function to create an invoice with a costumer and samples or pools."""
        invoice = store.invoice(invoice_id=invoice_id)
        if not invoice:
            customer_obj = StoreHelpers.ensure_customer(
                store=store,
                customer_id=customer_id,
            )

            user_obj: models.User = StoreHelpers.ensure_user(store=store, customer=customer_obj)
            customer_obj.invoice_contact: models.User = user_obj

            pool = []
            sample = []
            if record_type == "Sample":
                sample.append(
                    StoreHelpers.add_sample(
                        store=store, customer_id=customer_id, internal_id=internal_id
                    )
                )
                pool = None
            else:
                pool.append(
                    StoreHelpers.ensure_pool(store, customer_id=customer_id, name=customer_id)
                )
                sample = None

            invoice = store.add_invoice(
                customer=customer_obj,
                samples=sample,
                pools=pool,
                comment="just a test invoice",
                discount=discount,
                record_type=record_type,
            )
            store.add_commit(invoice)

        return invoice

    @classmethod
    def add_case_with_sample(cls, base_store: Store, case_id: str, sample_id: str) -> models.Family:
        """Helper function to add a case associated with a sample with the given ids."""

        case = cls.add_case(store=base_store, internal_id=case_id, name=case_id)
        sample = cls.add_sample(store=base_store, internal_id=sample_id)
        cls.add_relationship(store=base_store, sample=sample, case=case)
        return case



from typing import Optional, List

from sqlalchemy.orm import Query

from cg.constants.constants import SampleType
from cg.store.models import Sample


def get_sample_by_sample_id(internal_id: str, samples: Query, **kwargs) -> Query:
    """Return sample with sample id."""
    return samples.filter_by(internal_id=internal_id)


def get_samples_with_type(samples: Query, tissue_type: SampleType, **kwargs) -> Query:
    """Return samples with sample type."""
    is_tumour: bool = tissue_type == SampleType.TUMOR
    return samples.filter(Sample.is_tumour == is_tumour)


def get_samples_with_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Return samples with a loqusdb ID."""
    return samples.filter(Sample.loqusdb_id.isnot(None))


def get_samples_without_loqusdb_id(samples: Query, **kwargs) -> Query:
    """Return samples without a loqusdb ID."""
    return samples.filter(Sample.loqusdb_id.is_(None))


def get_sample_by_entry_id(entry_id: int, samples: Query, **kwargs) -> Query:
    """Return sample with entry id."""
    return samples.filter_by(id=entry_id)


def apply_sample_filter(
    functions: List[str],
    samples: Query,
    entry_id: Optional[int] = None,
    internal_id: Optional[str] = None,
    tissue_type: Optional[SampleType] = None,
) -> Query:
    """Apply filtering functions to the sample queries and return filtered results."""
    filter_map = {
        "get_sample_by_sample_id": get_sample_by_sample_id,
        "get_sample_by_entry_id": get_sample_by_entry_id,
        "get_samples_with_type": get_samples_with_type,
        "get_samples_with_loqusdb_id": get_samples_with_loqusdb_id,
        "get_samples_without_loqusdb_id": get_samples_without_loqusdb_id,
    }
    for function in functions:
        samples: Query = filter_map[function](
            entry_id=entry_id, internal_id=internal_id, samples=samples, tissue_type=tissue_type
        )
    return samples



"""
Code to handle communications to the shell from CG.
"""

import copy
import logging
import subprocess
from pathlib import Path
from subprocess import CalledProcessError
from typing import Dict, List

from cg.constants.process import RETURN_SUCCESS

LOG = logging.getLogger(__name__)


class Process:
    """Class to handle communication with other programs via the shell.

    The other parts of the code should not need to have any knowledge about how the processes are
    called, that will be handled in this module.Output form stdout and stdin will be handled here.
    """

    def __init__(
        self,
        binary: str,
        conda_binary: str = None,
        config: str = None,
        config_parameter: str = "--config",
        environment: str = None,
        launch_directory: str = None,
    ):
        """
        Args:
            binary(str): Path to binary for the process to use
            config(str): Path to config if used by process
            environment(str): Activate conda environment before executing binary
        """
        super(Process, self).__init__()
        self.binary: str = binary
        self.conda_binary: str = conda_binary
        self.config: str = config
        self.environment: str = environment
        self.launch_directory: str = launch_directory
        LOG.debug("Initialising Process with binary: %s", self.binary)
        self.base_call: List[str] = [self.binary]

        if conda_binary:
            LOG.debug(f"Activating environment with conda run for binary: {self.conda_binary}")
            self.base_call.insert(0, f"{self.conda_binary} run --name {self.environment}")
        elif environment:
            LOG.debug(f"Activating environment with: {self.environment}")
            self.base_call.insert(0, f"source activate {self.environment};")
        if config:
            self.base_call.extend([config_parameter, config])
        if launch_directory:
            self.base_call.insert(0, f"cd {self.launch_directory};")

        LOG.debug(f"Use base call {self.base_call}")
        self._stdout = ""
        self._stderr = ""

    def export_variables(self, export: Dict[str, str]) -> None:
        """Export variables prior to execution."""
        if export:
            self.base_call.insert(
                0, " ".join([f"export {variable}={value};" for variable, value in export.items()])
            )

    def run_command(self, parameters: list = None, dry_run: bool = False) -> int:
        """Execute a command in the shell.
        If environment is supplied - shell=True has to be supplied to enable passing as a string for executing multiple
         commands

        Args:
            parameters(list):
            dry_run(bool): Print command instead of executing it
        Return(int): Return code from called process

        """
        command = copy.deepcopy(self.base_call)
        if parameters:
            command.extend(parameters)

        LOG.info("Running command %s", " ".join(command))
        if dry_run:
            LOG.info("Dry run: process call will not be executed!!")
            return RETURN_SUCCESS

        if self.environment:
            res = subprocess.run(
                " ".join(command),
                shell=True,
                check=False,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        else:
            res = subprocess.run(
                command, check=False, stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )

        self.stdout = res.stdout.decode("utf-8").rstrip()
        self.stderr = res.stderr.decode("utf-8").rstrip()
        if res.returncode != RETURN_SUCCESS:
            LOG.critical("Call %s exit with a non zero exit code", command)
            LOG.critical(self.stderr)
            raise CalledProcessError(res.returncode, command)

        return res.returncode

    @property
    def stdout(self):
        """Fetch stdout"""
        return self._stdout

    @stdout.setter
    def stdout(self, text):
        self._stdout = text

    @stdout.deleter
    def stdout(self):
        del self._stdout

    @property
    def stderr(self):
        """Fetch stderr"""
        return self._stderr

    @stderr.setter
    def stderr(self, text):
        self._stderr = text

    @stderr.deleter
    def stderr(self):
        del self._stderr

    def stdout_lines(self):
        """Iterate over the lines in self.stdout"""
        yield from self.stdout.split("\n")

    def stderr_lines(self):
        """Iterate over the lines in self.stderr"""
        yield from self.stderr.split("\n")

    def __repr__(self):
        return f"Process:base_call:{self.base_call}"



from pathlib import Path

from cg.apps.cgstats.parsers.demux_stats import DemuxStats


def test_parse_demux_stats(demultiplexing_stats_path: Path):
    # GIVEN an existing demultiplexing stats file
    assert demultiplexing_stats_path.exists()
    # GIVEN that the file have some content
    demux_info: str = demultiplexing_stats_path.read_text()
    assert demux_info

    # WHEN instantiating a demux stats parser
    parser: DemuxStats = DemuxStats(demux_stats_path=demultiplexing_stats_path)

    # THEN assert that the parser have some content
    assert parser.lanes_to_barcode



"""Module for Rnafusion Analysis API."""

import logging
from pathlib import Path
from typing import Dict, List, Optional

from pydantic import ValidationError

from cg import resources
from cg.constants import Pipeline
from cg.constants.constants import FileFormat
from cg.constants.nextflow import NFX_READ1_HEADER, NFX_READ2_HEADER, NFX_SAMPLE_HEADER
from cg.constants.rnafusion import (
    RNAFUSION_SAMPLESHEET_HEADERS,
    RNAFUSION_STRANDEDNESS_HEADER,
    RnafusionDefaults,
)
from cg.io.controller import WriteFile
from cg.meta.workflow.analysis import AnalysisAPI
from cg.meta.workflow.fastq import RnafusionFastqHandler
from cg.meta.workflow.nextflow_common import NextflowAnalysisAPI
from cg.models.cg_config import CGConfig
from cg.models.nextflow.deliverables import NextflowDeliverables, replace_dict_values
from cg.models.rnafusion.rnafusion_sample import RnafusionSample
from cg.utils import Process

LOG = logging.getLogger(__name__)


class RnafusionAnalysisAPI(AnalysisAPI):
    """Handles communication between RNAFUSION processes
    and the rest of CG infrastructure."""

    def __init__(
        self,
        config: CGConfig,
        pipeline: Pipeline = Pipeline.RNAFUSION,
    ):
        super().__init__(config=config, pipeline=pipeline)
        self.root_dir: str = config.rnafusion.root
        self.nfcore_pipeline_path: str = config.rnafusion.pipeline_path
        self.references: str = config.rnafusion.references
        self.profile: str = config.rnafusion.profile
        self.conda_env: str = config.rnafusion.conda_env
        self.conda_binary: str = config.rnafusion.conda_binary

    @property
    def root(self) -> str:
        return self.root_dir

    @property
    def fastq_handler(self):
        return RnafusionFastqHandler

    @property
    def process(self):
        if not self._process:
            self._process = Process(
                binary=self.config.rnafusion.binary_path,
                environment=self.conda_env,
                conda_binary=self.conda_binary,
                launch_directory=self.config.rnafusion.launch_directory,
            )
        return self._process

    def get_profile(self, profile: Optional[str] = None) -> str:
        if profile:
            return profile
        return self.profile

    def get_case_config_path(self, case_id):
        return NextflowAnalysisAPI.get_case_config_path(case_id=case_id, root_dir=self.root_dir)

    def get_variables_to_export(self, case_id) -> Dict[str, str]:
        return NextflowAnalysisAPI.get_variables_to_export(case_id=case_id, root_dir=self.root_dir)

    def verify_analysis_finished(self, case_id):
        return NextflowAnalysisAPI.verify_analysis_finished(case_id=case_id, root_dir=self.root_dir)

    @staticmethod
    def build_samplesheet_content(
        case_id: str, fastq_r1: List[str], fastq_r2: List[str], strandedness: str
    ) -> Dict[str, List[str]]:
        """Build samplesheet headers and lists"""
        try:
            RnafusionSample(
                sample=case_id, fastq_r1=fastq_r1, fastq_r2=fastq_r2, strandedness=strandedness
            )
        except ValidationError as error:
            LOG.error(error)
            raise ValueError

        samples_full_list: list = []
        strandedness_full_list: list = []
        # Complete sample and strandedness lists to the same length as fastq_r1:
        for _ in range(len(fastq_r1)):
            samples_full_list.append(case_id)
            strandedness_full_list.append(strandedness)

        samplesheet_content: dict = {
            NFX_SAMPLE_HEADER: samples_full_list,
            NFX_READ1_HEADER: fastq_r1,
            NFX_READ2_HEADER: fastq_r2,
            RNAFUSION_STRANDEDNESS_HEADER: strandedness_full_list,
        }
        return samplesheet_content

    def write_samplesheet(self, case_id: str, strandedness: str, dry_run: bool = False) -> None:
        """Write sample sheet for rnafusion analysis in case folder."""
        case_obj = self.status_db.family(case_id)
        if len(case_obj.links) != 1:
            raise NotImplementedError(
                "Case objects are assumed to be related to a single sample (one link)"
            )

        for link in case_obj.links:
            sample_metadata: List[str] = self.gather_file_metadata_for_sample(link.sample)
            fastq_r1: List[str] = NextflowAnalysisAPI.extract_read_files(1, sample_metadata)
            fastq_r2: List[str] = NextflowAnalysisAPI.extract_read_files(2, sample_metadata)
            samplesheet_content: Dict[str, List[str]] = self.build_samplesheet_content(
                case_id, fastq_r1, fastq_r2, strandedness
            )
            LOG.info(samplesheet_content)
            if dry_run:
                continue
            NextflowAnalysisAPI.create_samplesheet_csv(
                samplesheet_content=samplesheet_content,
                headers=RNAFUSION_SAMPLESHEET_HEADERS,
                config_path=NextflowAnalysisAPI.get_case_config_path(
                    case_id=case_id, root_dir=self.root_dir
                ),
            )

    def write_params_file(self, case_id: str, dry_run: bool = False) -> None:
        """Write params-file for rnafusion analysis in case folder."""
        default_options: Dict[str, str] = self.get_verified_arguments(
            case_id=case_id,
            input=None,
            outdir=None,
            genomes_base=None,
            trim=RnafusionDefaults.TRIM,
            fusioninspector_filter=RnafusionDefaults.FUSIONINSPECTOR_FILTER,
            all=RnafusionDefaults.ALL,
            pizzly=RnafusionDefaults.PIZZLY,
            squid=RnafusionDefaults.SQUID,
            starfusion=RnafusionDefaults.STARFUSION,
            fusioncatcher=RnafusionDefaults.FUSIONCATCHER,
            arriba=RnafusionDefaults.ARRIBA,
            cli=False,
        )
        LOG.info(default_options)
        if dry_run:
            return
        NextflowAnalysisAPI.write_nextflow_yaml(
            content=default_options,
            file_path=NextflowAnalysisAPI.get_params_file_path(
                case_id=case_id, root_dir=self.root_dir
            ),
        )

    def get_references_path(self, genomes_base: Optional[Path] = None) -> Path:
        if genomes_base:
            return genomes_base
        return Path(self.references)

    def get_verified_arguments(
        self,
        case_id: str,
        input: Optional[str],
        outdir: Optional[str],
        genomes_base: Optional[str],
        trim: bool,
        fusioninspector_filter: bool,
        all: bool,
        pizzly: bool,
        squid: bool,
        starfusion: bool,
        fusioncatcher: bool,
        arriba: bool,
        cli: bool = False,
    ) -> Dict[str, str]:
        """Transforms click argument related to rnafusion that were left empty into
        defaults constructed with case_id paths or from config."""

        prefix: str = "--" if cli else ""

        return {
            f"{prefix}input": NextflowAnalysisAPI.get_input_path(
                case_id=case_id, root_dir=self.root_dir, input=input
            ).as_posix(),
            f"{prefix}outdir": NextflowAnalysisAPI.get_outdir_path(
                case_id=case_id, root_dir=self.root_dir, outdir=outdir
            ).as_posix(),
            f"{prefix}genomes_base": self.get_references_path(genomes_base=genomes_base).as_posix(),
            f"{prefix}trim": trim,
            f"{prefix}fusioninspector_filter": fusioninspector_filter,
            f"{prefix}all": all,
            f"{prefix}pizzly": pizzly,
            f"{prefix}squid": squid,
            f"{prefix}starfusion": starfusion,
            f"{prefix}fusioncatcher": fusioncatcher,
            f"{prefix}arriba": arriba,
        }

    @staticmethod
    def __build_command_str(options: dict, exclude_true: bool = False) -> List[str]:
        formatted_options: list = []
        for key, val in options.items():
            if val:
                if exclude_true and val is True:
                    formatted_options.append(str(key))
                elif val:
                    formatted_options.append(str(key))
                    formatted_options.append(str(val))
        return formatted_options

    def config_case(
        self,
        case_id: str,
        strandedness: str,
        dry_run: bool,
    ) -> None:
        """Create sample sheet file for RNAFUSION analysis."""
        NextflowAnalysisAPI.make_case_folder(
            case_id=case_id, root_dir=self.root_dir, dry_run=dry_run
        )
        LOG.info("Generating samplesheet")
        self.write_samplesheet(case_id=case_id, strandedness=strandedness, dry_run=dry_run)
        LOG.info("Generating parameters file")
        self.write_params_file(case_id=case_id, dry_run=dry_run)
        if dry_run:
            LOG.info("Dry run: Config files will not be written")
            return

        LOG.info("Configs files written")

    def run_analysis(
        self,
        case_id: str,
        log: Optional[str],
        work_dir: Optional[str],
        resume: bool,
        profile: str,
        with_tower: bool,
        stub: bool,
        config: Optional[str],
        params_file: Optional[str],
        dry_run: bool = False,
    ) -> None:
        """Execute RNAFUSION run analysis with given options."""

        run_options: List[str] = self.__build_command_str(
            options=NextflowAnalysisAPI.get_verified_arguments_run(
                case_id=case_id,
                root_dir=self.root_dir,
                work_dir=work_dir,
                resume=resume,
                profile=self.get_profile(profile=profile),
                with_tower=with_tower,
                stub=stub,
                params_file=params_file,
            ),
            exclude_true=True,
        )

        nextflow_options: List[str] = self.__build_command_str(
            options=NextflowAnalysisAPI.get_verified_arguments_nextflow(
                case_id=case_id,
                pipeline=self.pipeline,
                root_dir=self.root_dir,
                log=log,
                bg=True,
                quiet=True,
                config=config,
            ),
            exclude_true=True,
        )
        command: List[str] = ["run", self.nfcore_pipeline_path]
        parameters = (
            nextflow_options
            + command
            + run_options
            + NextflowAnalysisAPI.get_nextflow_stdout_stderr(
                case_id=case_id, root_dir=self.root_dir
            )
        )
        self.process.export_variables(
            export=self.get_variables_to_export(case_id),
        )
        self.process.run_command(parameters=parameters, dry_run=dry_run)

    def verify_case_config_file_exists(self, case_id: str) -> None:
        NextflowAnalysisAPI.verify_case_config_file_exists(case_id=case_id, root_dir=self.root_dir)

    def get_deliverables_file_path(self, case_id: str) -> Path:
        return NextflowAnalysisAPI.get_deliverables_file_path(
            case_id=case_id, root_dir=self.root_dir
        )

    def get_pipeline_version(self, case_id: str) -> str:
        return NextflowAnalysisAPI.get_pipeline_version(
            case_id=case_id, root_dir=self.root_dir, pipeline=self.pipeline
        )

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        NextflowAnalysisAPI.verify_deliverables_file_exists(case_id=case_id, root_dir=self.root_dir)

    def report_deliver(self, case_id: str) -> None:
        """Get a deliverables file template from resources, parse it and, then write the deliverables file."""
        deliverables_content: dict = NextflowAnalysisAPI.get_template_deliverables_file_content(
            resources.rnafusion_bundle_filenames_path
        )
        try:
            for index, deliver_file in enumerate(deliverables_content):
                NextflowDeliverables(deliverables=deliver_file)
                deliverables_content[index] = replace_dict_values(
                    NextflowAnalysisAPI.get_replace_map(case_id=case_id, root_dir=self.root_dir),
                    deliver_file,
                )
        except ValidationError as error:
            LOG.error(error)
            raise ValueError
        NextflowAnalysisAPI.make_case_folder(case_id=case_id, root_dir=self.root_dir)
        NextflowAnalysisAPI.write_deliverables_bundle(
            deliverables_content=NextflowAnalysisAPI.add_bundle_header(
                deliverables_content=deliverables_content
            ),
            file_path=NextflowAnalysisAPI.get_deliverables_file_path(
                case_id=case_id, root_dir=self.root_dir
            ),
        )
        LOG.info(
            "Writing deliverables file in "
            + str(
                NextflowAnalysisAPI.get_deliverables_file_path(
                    case_id=case_id, root_dir=self.root_dir
                )
            )
        )



import logging
from pathlib import Path

import cg
from cg.cli.base import base, init
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner, Result


def test_cli_version(cli_runner: CliRunner):
    # GIVEN I want to see the version of the program
    # WHEN asking to see the version
    result: Result = cli_runner.invoke(base, ["--version"])
    # THEN it should display the version of the program
    # THEN it should print the name and version of the tool only
    assert cg.__title__ in result.output
    assert cg.__version__ in result.output


def test_list_commands(cli_runner: CliRunner):
    # WHEN using simplest command 'cg'
    result = cli_runner.invoke(base, [])
    # THEN it should just work ;-)
    assert result.exit_code == 0


def test_missing_command(cli_runner: CliRunner):
    # WHEN invoking a missing command
    result = cli_runner.invoke(cg, ["i_dont_exist"])
    # THEN context should abort
    assert result.exit_code != 0


def test_cli_init(cli_runner: CliRunner, base_context: CGConfig, caplog):
    caplog.set_level(logging.INFO)
    # GIVEN you want to setup a new database using the CLI
    database = "./test_db.sqlite3"
    database_path = Path(database)
    database_uri = f"sqlite:///{database}"
    base_context.status_db_ = Store(uri=database_uri)
    with cli_runner.isolated_filesystem():
        assert database_path.exists() is False

        # WHEN calling "init"
        result = cli_runner.invoke(init, [], obj=base_context)

        # THEN it should setup the database with some tables
        assert result.exit_code == 0
        assert database_path.exists()
        assert len(Store(database_uri).engine.table_names()) > 0

        # GIVEN the database already exists
        # WHEN calling the init function
        result = cli_runner.invoke(init, [], obj=base_context)

        # THEN it should print an error and give error exit code
        assert result.exit_code != 0
        assert "Database already exists" in caplog.text

        # GIVEN the database already exists
        # WHEN calling "init" with "--reset"
        result = cli_runner.invoke(init, ["--reset"], input="Yes", obj=base_context)

        # THEN it should re-setup the tables and print new tables
        assert result.exit_code == 0
        assert "Success!" in caplog.text



from cg.utils.enums import StrEnum


class CustomerNames(StrEnum):
    cust999: str = "cust999"
    cust032: str = "cust032"
    cust001: str = "cust001"
    cust132: str = "cust132"


class CostCenters(StrEnum):
    ki: str = "ki"
    kth: str = "kth"



import coloredlogs
import requests
from flask import Flask, redirect, session, url_for
from flask_admin.base import AdminIndexView
from flask_dance.consumer import oauth_authorized
from flask_dance.contrib.google import google, make_google_blueprint

from cg.store import models

from . import admin, api, ext, invoices


def create_app():
    """Generate a flask application."""
    app = Flask(__name__, template_folder="templates")
    _load_config(app)
    _configure_extensions(app)
    _register_blueprints(app)

    return app


def _load_config(app):
    app.config.from_object(__name__.replace("app", "config"))


def _configure_extensions(app: Flask):
    _initialize_logging(app)
    certs_resp = requests.get("https://www.googleapis.com/oauth2/v1/certs")
    app.config["GOOGLE_OAUTH_CERTS"] = certs_resp.json()

    ext.cors.init_app(
        app, resources={app.config["CORS_RESOURCES"]: {"origins": app.config["CORS_ORIGINS"]}}
    )
    ext.csrf.init_app(app)
    ext.db.init_app(app)
    ext.lims.init_app(app)
    if app.config["OSTICKET_API_KEY"]:
        ext.osticket.init_app(app)
    ext.admin.init_app(app, index_view=AdminIndexView(endpoint="admin"))


def _initialize_logging(app):
    coloredlogs.install(level="DEBUG" if app.debug else "INFO")


def _register_blueprints(app: Flask):
    if not app.config["CG_ENABLE_ADMIN"]:
        return

    oauth_bp = make_google_blueprint(
        client_id=app.config["GOOGLE_OAUTH_CLIENT_ID"],
        client_secret=app.config["GOOGLE_OAUTH_CLIENT_SECRET"],
        scope=["openid", "https://www.googleapis.com/auth/userinfo.email"],
    )

    @oauth_authorized.connect_via(oauth_bp)
    def logged_in(blueprint, token):
        """Called when the user logs in via Google OAuth."""
        resp = google.get("/oauth2/v1/userinfo?alt=json")
        assert resp.ok, resp.text
        user_data = resp.json()
        session["user_email"] = user_data["email"]

    app.register_blueprint(api.BLUEPRINT)
    app.register_blueprint(invoices.BLUEPRINT, url_prefix="/invoices")
    app.register_blueprint(oauth_bp, url_prefix="/login")
    _register_admin_views()

    ext.csrf.exempt(api.BLUEPRINT)  # Protected with Auth header already

    @app.route("/")
    def index():
        return redirect(url_for("admin.index"))

    @app.route("/logout")
    def logout():
        """Log out the user."""
        session["user_email"] = None
        return redirect(url_for("index"))


def _register_admin_views():
    # Base data views
    ext.admin.add_view(admin.ApplicationView(models.Application, ext.db.session))
    ext.admin.add_view(admin.ApplicationVersionView(models.ApplicationVersion, ext.db.session))
    ext.admin.add_view(admin.BedView(models.Bed, ext.db.session))
    ext.admin.add_view(admin.BedVersionView(models.BedVersion, ext.db.session))
    ext.admin.add_view(admin.CustomerView(models.Customer, ext.db.session))
    ext.admin.add_view(admin.CollaborationView(models.Collaboration, ext.db.session))
    ext.admin.add_view(admin.OrganismView(models.Organism, ext.db.session))
    ext.admin.add_view(admin.PanelView(models.Panel, ext.db.session))
    ext.admin.add_view(admin.UserView(models.User, ext.db.session))

    # Business data views
    ext.admin.add_view(admin.FamilyView(models.Family, ext.db.session))
    ext.admin.add_view(admin.FamilySampleView(models.FamilySample, ext.db.session))
    ext.admin.add_view(admin.SampleView(models.Sample, ext.db.session))
    ext.admin.add_view(admin.PoolView(models.Pool, ext.db.session))
    ext.admin.add_view(admin.FlowcellView(models.Flowcell, ext.db.session))
    ext.admin.add_view(admin.AnalysisView(models.Analysis, ext.db.session))
    ext.admin.add_view(admin.DeliveryView(models.Delivery, ext.db.session))
    ext.admin.add_view(admin.InvoiceView(models.Invoice, ext.db.session))



import datetime as dt
import logging
from typing import List, Set

from cgmodels.cg.constants import Pipeline

from cg.constants import DataDelivery
from cg.constants.constants import CaseActions
from cg.exc import OrderError
from cg.meta.orders.lims import process_lims
from cg.meta.orders.submitter import Submitter
from cg.models.orders.order import OrderIn
from cg.models.orders.samples import Of1508Sample, OrderInSample
from cg.store import models

from cg.constants import Priority

LOG = logging.getLogger(__name__)


class CaseSubmitter(Submitter):
    def validate_order(self, order: OrderIn) -> None:
        self._validate_samples_available_to_customer(
            samples=order.samples, customer_id=order.customer
        )
        self._validate_case_names_are_unique(samples=order.samples, customer_id=order.customer)
        self._validate_subject_sex(samples=order.samples, customer_id=order.customer)

    def _validate_subject_sex(self, samples: [Of1508Sample], customer_id: str):
        """Validate that sex is consistent with existing samples, skips samples of unknown sex

        Args:
            samples     (list[dict]):   Samples to validate
            customer_id (str):          Customer that the samples belong to
        Returns:
            Nothing
        """

        sample: Of1508Sample
        for sample in samples:
            subject_id: str = sample.subject_id
            if not subject_id:
                continue
            new_gender: str = sample.sex
            if new_gender == "unknown":
                continue
            existing_samples: [models.Sample] = self.status.samples_by_subject_id(
                customer_id=customer_id, subject_id=subject_id
            )
            existing_sample: models.Sample
            for existing_sample in existing_samples:
                previous_gender = existing_sample.sex
                if previous_gender == "unknown":
                    continue

                if previous_gender != new_gender:
                    raise OrderError(
                        f"Sample gender inconsistency for subject_id: {subject_id}: previous gender {previous_gender}, new gender {new_gender}"
                    )

    def _validate_samples_available_to_customer(
        self, samples: List[OrderInSample], customer_id: str
    ) -> None:
        """Validate that the customer have access to all samples"""
        sample: Of1508Sample
        for sample in samples:
            if not sample.internal_id:
                continue

            existing_sample: models.Sample = self.status.sample(sample.internal_id)
            data_customer: models.Customer = self.status.customer(customer_id)

            if existing_sample.customer not in data_customer.collaborators:
                raise OrderError(f"Sample not available: {sample.name}")

    def _validate_case_names_are_unique(
        self, samples: List[OrderInSample], customer_id: str
    ) -> None:
        """Validate that the names of all cases are unused for all samples"""
        customer_obj: models.Customer = self.status.customer(customer_id)

        sample: Of1508Sample
        for sample in samples:
            if self._is_rerun_of_existing_case(sample):
                continue

            if self.status.find_family(customer=customer_obj, name=sample.family_name):
                raise OrderError(f"Case name {sample.family_name} already in use")

    def submit_order(self, order: OrderIn) -> dict:
        """Submit a batch of samples for sequencing and analysis."""
        result = self._process_case_samples(order=order)
        for case_obj in result["records"]:
            LOG.info(f"{case_obj.name}: submit family samples")
            status_samples = [
                link_obj.sample
                for link_obj in case_obj.links
                if link_obj.sample.original_ticket == order.ticket
            ]
            self._add_missing_reads(status_samples)
        return result

    def _process_case_samples(self, order: OrderIn) -> dict:
        """Process samples to be analyzed."""
        project_data = lims_map = None

        # submit new samples to lims
        new_samples = [sample for sample in order.samples if sample.internal_id is None]
        if new_samples:
            project_data, lims_map = process_lims(
                lims_api=self.lims, lims_order=order, new_samples=new_samples
            )

        status_data = self.order_to_status(order=order)
        samples = [sample for family in status_data["families"] for sample in family["samples"]]
        if lims_map:
            self._fill_in_sample_ids(samples, lims_map)

        new_families = self.store_items_in_status(
            customer=status_data["customer"],
            order=status_data["order"],
            ordered=project_data["date"] if project_data else dt.datetime.now(),
            ticket=order.ticket,
            items=status_data["families"],
        )
        return {"project": project_data, "records": new_families}

    @staticmethod
    def _group_cases(samples: List[Of1508Sample]) -> dict:
        """Group samples in cases."""
        cases = {}
        for sample in samples:
            case_id = sample.family_name
            if case_id not in cases:
                cases[case_id] = []
            cases[case_id].append(sample)
        return cases

    @staticmethod
    def _get_single_value(case_name, case_samples, value_key, value_default=None):
        values = set(getattr(sample, value_key) or value_default for sample in case_samples)
        if len(values) > 1:
            raise ValueError(f"different sample {value_key} values: {case_name} - {values}")
        single_value = values.pop()
        return single_value

    @staticmethod
    def order_to_status(order: OrderIn) -> dict:
        """Converts order input to status interface input for MIP-DNA, MIP-RNA and Balsamic."""
        status_data = {"customer": order.customer, "order": order.name, "families": []}
        cases = CaseSubmitter._group_cases(order.samples)

        for case_name, case_samples in cases.items():
            case_internal_id: str = CaseSubmitter._get_single_value(
                case_name, case_samples, "case_internal_id"
            )
            cohorts: Set[str] = {
                cohort for sample in case_samples for cohort in sample.cohorts if cohort
            }
            data_analysis = CaseSubmitter._get_single_value(
                case_name, case_samples, "data_analysis"
            )
            data_delivery = CaseSubmitter._get_single_value(
                case_name, case_samples, "data_delivery"
            )

            panels: Set[str] = set()
            if data_analysis == Pipeline.MIP_DNA:
                panels: Set[str] = {
                    panel for sample in case_samples for panel in sample.panels if panel
                }

            priority = CaseSubmitter._get_single_value(
                case_name, case_samples, "priority", Priority.standard.name
            )
            synopsis: str = CaseSubmitter._get_single_value(case_name, case_samples, "synopsis")

            case = {
                "cohorts": list(cohorts),
                "data_analysis": data_analysis,
                "data_delivery": data_delivery,
                "internal_id": case_internal_id,
                "name": case_name,
                "panels": list(panels),
                "priority": priority,
                "samples": [
                    {
                        "age_at_sampling": sample.age_at_sampling,
                        "application": sample.application,
                        "capture_kit": sample.capture_kit,
                        "comment": sample.comment,
                        "control": sample.control,
                        "father": sample.father,
                        "internal_id": sample.internal_id,
                        "mother": sample.mother,
                        "name": sample.name,
                        "phenotype_groups": list(sample.phenotype_groups),
                        "phenotype_terms": list(sample.phenotype_terms),
                        "reference_genome": sample.reference_genome
                        if hasattr(sample, "reference_genome")
                        else None,
                        "sex": sample.sex,
                        "status": sample.status if hasattr(sample, "status") else None,
                        "subject_id": sample.subject_id,
                        "tumour": sample.tumour,
                    }
                    for sample in case_samples
                ],
                "synopsis": synopsis,
            }

            status_data["families"].append(case)
        return status_data

    def store_items_in_status(
        self, customer: str, order: str, ordered: dt.datetime, ticket: str, items: List[dict]
    ) -> List[models.Family]:
        """Store cases and samples in the status database."""

        customer_obj = self.status.customer(customer)
        new_families = []
        for case in items:
            case_obj = self.status.family(case["internal_id"])
            if not case_obj:
                case_obj = self._create_case(case, customer_obj, ticket)
                new_families.append(case_obj)
            else:
                self._append_ticket(ticket=ticket, case=case_obj)
                self._update_action(action=CaseActions.ANALYZE, case=case_obj)

            self._update_case(case, case_obj)

            family_samples = {}
            for sample in case["samples"]:
                sample_obj = self.status.sample(sample["internal_id"])
                if not sample_obj:
                    sample_obj = self._create_sample(
                        case, customer_obj, order, ordered, sample, ticket
                    )

                family_samples[sample["name"]] = sample_obj

            for sample in case["samples"]:
                mother_obj = family_samples.get(sample.get("mother"))
                father_obj = family_samples.get(sample.get("father"))
                with self.status.session.no_autoflush:
                    link_obj = self.status.link(case_obj.internal_id, sample["internal_id"])
                if not link_obj:
                    link_obj = self._create_link(
                        case_obj, family_samples, father_obj, mother_obj, sample
                    )

                self._update_relationship(father_obj, link_obj, mother_obj, sample)

            self.status.add_commit(new_families)
        return new_families

    @staticmethod
    def _update_case(case, case_obj):
        case_obj.panels = case["panels"]

    @staticmethod
    def _append_ticket(ticket: str, case: models.Family) -> None:
        """Add a ticket to the case."""
        case.tickets = f"{case.tickets},{ticket}"

    @staticmethod
    def _update_action(action: str, case: models.Family) -> None:
        """Update action of a case."""
        case.action = action

    @staticmethod
    def _update_relationship(father_obj, link_obj, mother_obj, sample):
        link_obj.status = sample["status"] or link_obj.status
        link_obj.mother = mother_obj or link_obj.mother
        link_obj.father = father_obj or link_obj.father

    def _create_link(self, case_obj, family_samples, father_obj, mother_obj, sample):
        link_obj = self.status.relate_sample(
            family=case_obj,
            sample=family_samples[sample["name"]],
            status=sample["status"],
            mother=mother_obj,
            father=father_obj,
        )
        self.status.add(link_obj)
        return link_obj

    def _create_sample(self, case, customer_obj, order, ordered, sample, ticket):
        sample_obj = self.status.add_sample(
            name=sample["name"],
            comment=sample["comment"],
            control=sample["control"],
            internal_id=sample["internal_id"],
            order=order,
            ordered=ordered,
            original_ticket=ticket,
            tumour=sample["tumour"],
            age_at_sampling=sample["age_at_sampling"],
            capture_kit=sample["capture_kit"],
            phenotype_groups=sample["phenotype_groups"],
            phenotype_terms=sample["phenotype_terms"],
            priority=case["priority"],
            reference_genome=sample["reference_genome"],
            sex=sample["sex"],
            subject_id=sample["subject_id"],
        )
        sample_obj.customer = customer_obj
        with self.status.session.no_autoflush:
            application_tag = sample["application"]
            sample_obj.application_version = self.status.current_application_version(
                application_tag
            )
        self.status.add(sample_obj)
        new_delivery = self.status.add_delivery(destination="caesar", sample=sample_obj)
        self.status.add(new_delivery)
        return sample_obj

    def _create_case(self, case: dict, customer_obj: models.Customer, ticket: str):
        case_obj = self.status.add_case(
            cohorts=case["cohorts"],
            data_analysis=Pipeline(case["data_analysis"]),
            data_delivery=DataDelivery(case["data_delivery"]),
            name=case["name"],
            priority=case["priority"],
            synopsis=case["synopsis"],
            ticket=ticket,
        )
        case_obj.customer = customer_obj
        return case_obj

    @staticmethod
    def _is_rerun_of_existing_case(sample: Of1508Sample) -> bool:
        return sample.case_internal_id is not None



from enum import Enum
from typing import Optional, List, Union

from sqlalchemy.orm import Query

from cg.store.models import Flowcell, FamilySample, Family


class FlowCellFilters(str, Enum):
    get_flow_cells_by_case = ("get_flow_cells_by_case",)
    get_flow_cell_by_id = ("get_flow_cell_by_id",)
    get_flow_cell_by_id_and_by_enquiry = ("get_flow_cell_by_id_and_by_enquiry",)
    get_flow_cells_with_statuses = ("get_flow_cells_with_statuses",)


def get_flow_cells_by_case(case: Family, flow_cells: Query, **kwargs) -> Query:
    """Return flow cells by case id."""
    return flow_cells.filter(FamilySample.family == case)


def get_flow_cell_by_id(flow_cells: Query, flow_cell_id: str, **kwargs) -> Flowcell:
    """Return flow cell by flow cell id."""
    return flow_cells.filter(Flowcell.name == flow_cell_id).first()


def get_flow_cell_by_id_and_by_enquiry(flow_cells: Query, flow_cell_id: str, **kwargs) -> Query:
    """Return flow cell by flow cell id enquiry."""
    return flow_cells.filter(Flowcell.name.like(f"%{flow_cell_id}%"))


def get_flow_cells_with_statuses(
    flow_cells: Query, flow_cell_statuses: List[str], **kwargs
) -> Query:
    """Return flow cells by flow cell statuses."""
    return flow_cells.filter(Flowcell.status.in_(flow_cell_statuses))


def apply_flow_cell_filter(
    flow_cells: Query,
    functions: List[str],
    case: Optional[Family] = None,
    flow_cell_id: Optional[str] = None,
    flow_cell_statuses: Optional[List[str]] = None,
) -> Union[Query, Flowcell]:
    """Apply filtering functions and return filtered results."""
    filter_map = {
        FlowCellFilters.get_flow_cells_by_case: get_flow_cells_by_case,
        FlowCellFilters.get_flow_cell_by_id: get_flow_cell_by_id,
        FlowCellFilters.get_flow_cell_by_id_and_by_enquiry: get_flow_cell_by_id_and_by_enquiry,
        FlowCellFilters.get_flow_cells_with_statuses: get_flow_cells_with_statuses,
    }
    for function in functions:
        flow_cells: Union[Query, Flowcell] = filter_map[function](
            flow_cells=flow_cells,
            case=case,
            flow_cell_id=flow_cell_id,
            flow_cell_statuses=flow_cell_statuses,
        )
    return flow_cells



"""Transfer CLI."""
import logging
import click

from pathlib import Path

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.meta.transfer import PoolState, SampleState, TransferFlowCell, TransferLims
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Flowcell

LOG = logging.getLogger(__name__)


@click.group(name="transfer")
@click.pass_obj
def transfer_group(context: CGConfig):
    """Transfer results to the status interface."""
    lims_api: LimsAPI = context.lims_api
    status_db: Store = context.status_db
    hk_api: HousekeeperAPI = context.housekeeper_api
    stats_api: StatsAPI = context.cg_stats_api
    context.meta_apis["transfer_flow_cell_api"] = TransferFlowCell(
        db=status_db, stats_api=stats_api, hk_api=hk_api
    )
    context.meta_apis["transfer_lims_api"] = TransferLims(status=status_db, lims=lims_api)


@transfer_group.command("flow-cell")
@click.argument("flow-cell-id")
@click.option(
    "-d",
    "--flow-cell-dir",
    type=click.Path(exists=True, file_okay=False),
    required=True,
    help="Path to demultiplexed flow cell output directory",
)
@click.option(
    "--store/--no-store", default=True, help="Store sample bundles of flow cell in Housekeeper"
)
@click.pass_obj
def flow_cell(context: CGConfig, flow_cell_dir: str, flow_cell_id: str, store: bool = True):
    """Populate results from a flow cell."""
    flow_cell_dir: Path = Path(flow_cell_dir)
    status_db: Store = context.status_db
    transfer_api: TransferFlowCell = context.meta_apis["transfer_flow_cell_api"]
    new_record: Flowcell = transfer_api.transfer(
        flow_cell_dir=flow_cell_dir, flow_cell_id=flow_cell_id, store=store
    )
    status_db.add_commit(new_record)
    LOG.info(f"flow cell added: {new_record}")


@transfer_group.command()
@click.option(
    "-s", "--status", type=click.Choice(["received", "prepared", "delivered"]), default="received"
)
@click.option(
    "-i", "--include", type=click.Choice(["unset", "not-invoiced", "all"]), default="unset"
)
@click.option("--sample-id", help="Lims Submitted Sample id. use together with status.")
@click.pass_obj
def lims(context: CGConfig, status: str, include: str, sample_id: str):
    """Check if samples have been updated in LIMS."""
    transfer_api: TransferLims = context.meta_apis["transfer_lims_api"]
    transfer_api.transfer_samples(
        status_type=SampleState[status.upper()], include=include, sample_id=sample_id
    )


@transfer_group.command()
@click.option("-s", "--status", type=click.Choice(["received", "delivered"]), default="delivered")
@click.pass_obj
def pools(context: CGConfig, status: str):
    """
    Update pools with received_at or delivered_at dates from LIMS. Defaults to delivered if no
    option is provided.
    """
    transfer_api: TransferLims = context.meta_apis["transfer_lims_api"]
    transfer_api.transfer_pools(status_type=PoolState[status.upper()])



from typing import Iterable, List

import click
from cg.constants import CASE_ACTIONS, Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from ansi.colour import fg
from ansi.colour.fx import reset
from tabulate import tabulate

from cg.constants import Priority

from cg.utils.click.EnumChoice import EnumChoice

STATUS_OPTIONS = ["pending", "running", "completed", "failed", "error"]
CASE_HEADERS_LONG = [
    "Case",
    "Workflow",
    "Ordered",
    "Received",
    "Prepared",
    "Sequenced",
    "Flowcells",
    "Analysed",
    "Uploaded",
    "Delivered",
    "Delivery Reported",
    "Invoiced",
    "TAT",
]
ALWAYS_LONG_HEADERS = [
    CASE_HEADERS_LONG[0],
    CASE_HEADERS_LONG[1],
    CASE_HEADERS_LONG[6],
    CASE_HEADERS_LONG[7],
]
CASE_HEADERS_MEDIUM = []
CASE_HEADERS_SHORT = []

for header in CASE_HEADERS_LONG:
    if header not in ALWAYS_LONG_HEADERS:
        header = header[:3]

    CASE_HEADERS_MEDIUM.append(header)

    if header not in ALWAYS_LONG_HEADERS:
        header = header[0]

    CASE_HEADERS_SHORT.append(header)


@click.group()
def status():
    """View status of things."""
    pass


@status.command()
@click.pass_obj
def analysis(context: CGConfig):
    """Which families will be analyzed?"""
    records: List[models.Family] = context.status_db.cases_to_analyze(pipeline=Pipeline.MIP_DNA)
    for case_obj in records:
        click.echo(case_obj)


def present_bool(a_dict, param, show_false=False):
    """presents boolean value in a human friendly format"""
    value = a_dict.get(param)

    if show_false:
        return (
            "-"
            if value is None
            else "✓"
            if value is True
            else "✗"
            if value is False
            else str(value)
        )

    return "-" if value is None else "✓" if value is True else "" if value is False else str(value)


def present_date(a_dict, param, show_negative, show_time):
    """presents datetime value in a human friendly format"""
    value = a_dict.get(param)

    if not show_time and value and value.date:
        value = value.date()

    if show_negative:
        return str(value)

    return "" if not value else value or str(value)


def present_string(a_dict, param, show_negative):
    """presents string value in a human friendly format"""
    value = a_dict.get(param)

    if show_negative:
        return str(value)

    return "" if not value else value or str(value)


@status.command()
@click.pass_obj
@click.option(
    "-o",
    "--output",
    "output_type",
    type=click.Choice(["bool", "count", "date", "datetime"]),
    default="bool",
    help="how to display status",
)
@click.option("--verbose", is_flag=True, help="show status information otherwise left out")
@click.option("--days", default=31, help="days to go back")
@click.option("--internal-id", help="search by internal id")
@click.option("--name", help="search by name given by customer")
@click.option("--case-action", type=click.Choice(CASE_ACTIONS), help="filter by case action")
@click.option("--priority", type=EnumChoice(Priority, use_value=False), help="filter by priority")
@click.option("--data-analysis", help="filter on case data_analysis")
@click.option("--sample-id", help="filter by sample id")
@click.option("-c", "--customer-id", help="filter by customer")
@click.option("-C", "--exclude-customer-id", help="exclude customer")
@click.option("-r", "--only-received", is_flag=True, help="only completely received cases")
@click.option("-R", "--exclude-received", is_flag=True, help="exclude completely received cases")
@click.option("-p", "--only-prepared", is_flag=True, help="only completely prepared cases")
@click.option("-P", "--exclude-prepared", is_flag=True, help="exclude completely prepared cases")
@click.option("-s", "--only-sequenced", is_flag=True, help="only completely sequenced cases")
@click.option("-S", "--exclude-sequenced", is_flag=True, help="exclude completely sequenced cases")
@click.option("-a", "--only-analysed", is_flag=True, help="only analysed cases")
@click.option("-A", "--exclude-analysed", is_flag=True, help="exclude analysed cases")
@click.option("-u", "--only-uploaded", is_flag=True, help="only uploaded cases")
@click.option("-U", "--exclude-uploaded", is_flag=True, help="exclude uploaded cases")
@click.option("-d", "--only-delivered", is_flag=True, help="only LIMS delivered cases")
@click.option("-D", "--exclude-delivered", is_flag=True, help="exclude LIMS delivered cases")
@click.option("--dr", "--only-delivery-reported", is_flag=True, help="only delivery reported cases")
@click.option(
    "--DR", "--exclude-delivery-reported", is_flag=True, help="exclude delivery " "reported cases"
)
@click.option("-i", "--only-invoiced", is_flag=True, help="only completely invoiced cases")
@click.option("-I", "--exclude-invoiced", is_flag=True, help="exclude completely invoiced cases")
def cases(
    context: CGConfig,
    output_type,
    verbose,
    days,
    internal_id,
    name,
    case_action,
    priority,
    customer_id,
    data_analysis,
    sample_id,
    only_received,
    only_prepared,
    only_sequenced,
    only_analysed,
    only_uploaded,
    only_delivered,
    only_delivery_reported,
    only_invoiced,
    exclude_customer_id,
    exclude_received,
    exclude_prepared,
    exclude_sequenced,
    exclude_analysed,
    exclude_uploaded,
    exclude_delivered,
    exclude_delivery_reported,
    exclude_invoiced,
):
    """progress of each case"""
    status_db: Store = context.status_db
    records: List[models.Family] = status_db.cases(
        days=days,
        internal_id=internal_id,
        name=name,
        case_action=case_action,
        priority=priority,
        customer_id=customer_id,
        exclude_customer_id=exclude_customer_id,
        data_analysis=data_analysis,
        sample_id=sample_id,
        only_received=only_received,
        only_prepared=only_prepared,
        only_sequenced=only_sequenced,
        only_analysed=only_analysed,
        only_uploaded=only_uploaded,
        only_delivered=only_delivered,
        only_delivery_reported=only_delivery_reported,
        only_invoiced=only_invoiced,
        exclude_received=exclude_received,
        exclude_prepared=exclude_prepared,
        exclude_sequenced=exclude_sequenced,
        exclude_analysed=exclude_analysed,
        exclude_uploaded=exclude_uploaded,
        exclude_delivered=exclude_delivered,
        exclude_delivery_reported=exclude_delivery_reported,
        exclude_invoiced=exclude_invoiced,
    )
    case_rows = []

    if output_type == "bool":
        case_header = CASE_HEADERS_SHORT

    elif output_type == "count":
        case_header = CASE_HEADERS_MEDIUM

    elif output_type in ("date", "datetime"):
        case_header = CASE_HEADERS_LONG

    for case in records:
        tat_number = case.get("tat")
        max_tat = case.get("max_tat")

        if (
            case.get("samples_received_bool")
            and case.get("samples_delivered_bool")
            and tat_number <= max_tat
        ):
            tat_color = fg.green
        elif tat_number == max_tat:
            tat_color = fg.yellow
        elif tat_number > max_tat:
            tat_color = fg.red
        else:
            tat_color = fg.white

        color_start = tat_color
        color_end = str(reset)

        title = color_start + f"{case.get('internal_id')}"

        if name:
            title = f"{title} ({case.get('name')})"

        if (
            not case.get("case_external_bool")
            and case.get("samples_received_bool")
            and case.get("samples_delivered_bool")
        ):
            tat = f"{tat_number}/{max_tat}" + color_end
        elif case.get("case_external_bool") and case.get("analysis_uploaded_bool"):
            tat = f"{tat_number}/{max_tat}" + color_end
        else:
            tat = f"({tat_number})/{max_tat}" + color_end

        data_analysis = f"{case.get('data_analysis')}"

        show_time = output_type == "datetime"

        ordered = present_date(case, "ordered_at", verbose, show_time)

        if output_type == "bool":
            received = present_bool(case, "samples_received_bool", verbose)
            prepared = present_bool(case, "samples_prepared_bool", verbose)
            sequenced = present_bool(case, "samples_sequenced_bool", verbose)
            flowcell = present_bool(case, "flowcells_on_disk_bool", verbose)
            analysed_bool = present_bool(case, "analysis_completed_bool", verbose)

            if case.get("analysis_completed_at"):
                analysed = f"{analysed_bool}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{analysed_bool}{present_string(case, 'analysis_status',verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{analysed_bool}{present_string(case, 'case_action', verbose)}"

            uploaded = present_bool(case, "analysis_uploaded_bool", verbose)
            delivered = present_bool(case, "samples_delivered_bool", verbose)
            delivery_reported = present_bool(case, "analysis_delivery_reported_bool", verbose)
            invoiced = present_bool(case, "samples_invoiced_bool", verbose)

        elif output_type == "count":
            received = f"{case.get('samples_received')}/{case.get('samples_to_receive')}"
            prepared = f"{case.get('samples_prepared')}/{case.get('samples_to_prepare')}"
            sequenced = f"{case.get('samples_sequenced')}/{case.get('samples_to_sequence')}"
            flowcell = f"{case.get('flowcells_on_disk')}/{case.get('total_samples')}"

            if case.get("analysis_completed_at"):
                analysed = f"{present_date(case, 'analysis_completed_at', verbose, show_time)}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{present_string(case, 'analysis_status', verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{present_string(case, 'case_action', verbose)}"

            uploaded = present_date(case, "analysis_uploaded_at", verbose, show_time)
            delivered = f"{case.get('samples_delivered')}/{case.get('samples_to_deliver')}"
            delivery_reported = present_date(
                case, "analysis_delivery_reported_at", verbose, show_time
            )
            invoiced = f"{case.get('samples_invoiced')}/{case.get('samples_to_invoice')}"

        elif output_type in ("date", "datetime"):
            received = present_date(case, "samples_received_at", verbose, show_time)
            prepared = present_date(case, "samples_prepared_at", verbose, show_time)
            sequenced = present_date(case, "samples_sequenced_at", verbose, show_time)
            flowcell = present_string(case, "flowcells_status", verbose)

            if case.get("analysis_completed_at"):
                analysed = f"{present_date(case, 'analysis_completed_at', verbose, show_time)}"
            elif case.get("analysis_status"):
                analysed = (
                    f"{present_string(case, 'analysis_status', verbose)}"
                    f" {case.get('analysis_completion')}%"
                )
            else:
                analysed = f"{present_string(case, 'case_action', verbose)}"

            uploaded = present_date(case, "analysis_uploaded_at", verbose, show_time)
            delivered = present_date(case, "samples_delivered_at", verbose, show_time)
            delivery_reported = present_date(
                case, "analysis_delivery_reported_at", verbose, show_time
            )
            invoiced = present_date(case, "samples_invoiced_at", verbose, show_time)

        case_row = [
            title,
            data_analysis,
            ordered,
            received,
            prepared,
            sequenced,
            flowcell,
            analysed,
            uploaded,
            delivered,
            delivery_reported,
            invoiced,
            tat,
        ]
        case_rows.append(case_row)

    click.echo(tabulate(case_rows, headers=case_header, tablefmt="psql"))

    header_description = ""
    for i, _ in enumerate(case_header):
        if case_header[i] != CASE_HEADERS_LONG[i]:
            header_description = f"{header_description} {case_header[i]}={CASE_HEADERS_LONG[i]}"
    click.echo(header_description)


@status.command()
@click.option("-s", "--skip", default=0, help="skip initial records")
@click.pass_obj
def samples(context: CGConfig, skip: int):
    """View status of samples."""
    status_db: Store = context.status_db
    records: Iterable[models.Sample] = status_db.samples().offset(skip).limit(30)
    for record in records:
        message = f"{record.internal_id} ({record.customer.internal_id})"
        if record.sequenced_at:
            color = "green"
            message += f" [SEQUENCED: {record.sequenced_at.date()}]"
        elif record.received_at and record.reads:
            color = "orange"
            message += f" [READS: {record.reads}]"
        elif record.received_at:
            color = "blue"
            message += f" [RECEIVED: {record.received_at.date()}]"
        else:
            color = "white"
            message += " [NOT RECEIVED]"
        click.echo(click.style(message, fg=color))


@status.command()
@click.option("-s", "--skip", default=0, help="skip initial records")
@click.pass_obj
def families(context: CGConfig, skip: int):
    """View status of families."""
    click.echo("red: prio > 1, blue: prio = 1, green: completed, yellow: action")
    status_db: Store = context.status_db
    records: List[models.Family] = status_db.families().offset(skip).limit(30)
    for case_obj in records:
        color = "red" if case_obj.priority_int > 1 else "blue"
        message = f"{case_obj.internal_id} ({case_obj.priority_int})"
        if case_obj.analyses:
            message += f" {case_obj.analyses[0].completed_at.date()}"
            color = "green"
        if case_obj.action:
            message += f" [{case_obj.action.upper()}]"
            color = "yellow"
        click.echo(click.style(message, fg=color))



import logging

from click.testing import CliRunner

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.cli.store.fastq import (
    store_case,
    store_demultiplexed_flow_cell,
    store_flow_cell,
    store_sample,
    store_ticket,
)
from cg.constants import EXIT_SUCCESS
from cg.meta.compress import CompressAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from cg.store.models import Sample


def test_store_sample_when_no_samples(
    caplog, cli_runner: CliRunner, compress_context: CGConfig, sample_id: str
):
    """Test to run store samples command with a database without samples."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context without samples

    # WHEN running the store sample command
    res = cli_runner.invoke(store_sample, [sample_id], obj=compress_context)

    # THEN assert the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that sample cannot be found
    assert f"Could not find {sample_id}" in caplog.text


def test_store_sample_when_decompression_not_finished(
    cli_runner: CliRunner, caplog, mocker, populated_compress_context: CGConfig, sample_id: str
):
    """Test to run store samples command when decompression is not finished."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a sample

    # GIVEN that decompression is not finished
    mocker.patch.object(CompressAPI, "add_decompressed_fastq")
    CompressAPI.add_decompressed_fastq.return_value = False

    # WHEN running the store samples command
    res = cli_runner.invoke(store_sample, [sample_id], obj=populated_compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we skip the sample
    assert f"Skipping sample {sample_id}" in caplog.text


def test_store_sample_when_decompression_finished(
    caplog, cli_runner: CliRunner, mocker, populated_compress_context: CGConfig, sample_id: str
):
    """Test to run store samples command when decompression is finished."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a sample

    # GIVEN that decompression is finished
    mocker.patch.object(CompressAPI, "add_decompressed_fastq")
    CompressAPI.add_decompressed_fastq.return_value = True

    # WHEN running the store sample command
    res = cli_runner.invoke(store_sample, [sample_id], obj=populated_compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we stored FASTQ files
    assert f"Stored fastq files for {sample_id}" in caplog.text


def test_store_case_when_no_case(
    caplog, case_id: str, cli_runner: CliRunner, compress_context: CGConfig
):
    """Test to run store case command when no case in database."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with no case

    # WHEN running the store case command
    res = cli_runner.invoke(store_case, [case_id], obj=compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we cannot find the case
    assert f"Could not find case {case_id}" in caplog.text


def test_store_case(
    caplog,
    cli_runner: CliRunner,
    case_id: str,
    mocker,
    populated_compress_context: CGConfig,
    sample_id: str,
):
    """Test to run store case command."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a case and a sample

    # GIVEN that decompression is not finished
    mocker.patch.object(CompressAPI, "add_decompressed_fastq")
    CompressAPI.add_decompressed_fastq.return_value = True

    # WHEN running the store case command
    res = cli_runner.invoke(store_case, [case_id], obj=populated_compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we stored FASTQ files
    assert f"Stored fastq files for {sample_id}" in caplog.text


def test_store_flow_cell(
    caplog,
    cli_runner: CliRunner,
    flow_cell_id: str,
    mocker,
    populated_compress_context: CGConfig,
    sample_id: str,
):
    """Test to run store flow cell command."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a sample
    sample: Sample = populated_compress_context.status_db.sample(sample_id)

    # GIVEN samples objects on a flow cell
    mocker.patch.object(Store, "get_samples_from_flow_cell")
    Store.get_samples_from_flow_cell.return_value = [sample]

    # GIVEN that decompression is not finished
    mocker.patch.object(CompressAPI, "add_decompressed_fastq")
    CompressAPI.add_decompressed_fastq.return_value = True

    # WHEN running the store flow cell command
    res = cli_runner.invoke(store_flow_cell, [flow_cell_id], obj=populated_compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we stored FASTQ files
    assert f"Stored fastq files for {sample_id}" in caplog.text


def test_store_ticket(
    caplog,
    cli_runner: CliRunner,
    mocker,
    populated_compress_context: CGConfig,
    sample_id: str,
    ticket: str,
):
    """Test to run store ticket command."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a sample

    # GIVEN that decompression is not finished
    mocker.patch.object(CompressAPI, "add_decompressed_fastq")
    CompressAPI.add_decompressed_fastq.return_value = True

    # WHEN running the store ticket command
    res = cli_runner.invoke(store_ticket, [ticket], obj=populated_compress_context)

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS

    # THEN assert that we log that we stored FASTQ files
    assert f"Stored fastq files for {sample_id}" in caplog.text


def test_store_store_demultiplexed_flow_cell(
    caplog,
    cli_runner: CliRunner,
    flow_cell_id: str,
    helpers,
    mocker,
    real_populated_compress_context: CGConfig,
    sample_id: str,
):
    """Test to run store demultiplexed flow cell command."""
    caplog.set_level(logging.DEBUG)
    # GIVEN a context with a sample
    sample: Sample = real_populated_compress_context.status_db.sample(sample_id)

    # GIVEN samples objects on a flow cell
    mocker.patch.object(Store, "get_samples_from_flow_cell")
    Store.get_samples_from_flow_cell.return_value = [sample]

    # GIVEN an updated metadata file
    mocker.patch("cg.cli.store.fastq.update_metadata_paths", return_value=None)

    # WHEN running the store demultiplexed flow cell command
    res = cli_runner.invoke(
        store_demultiplexed_flow_cell, [flow_cell_id], obj=real_populated_compress_context
    )

    # THEN assert that the command exits successfully
    assert res.exit_code == EXIT_SUCCESS
